"""AutoUplift."""
import abc
import logging
import time

from abc import abstractmethod
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from dataclasses import field
from itertools import product
from itertools import zip_longest
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from pandas import DataFrame
from sklearn.model_selection import train_test_split

from lightautoml.addons.uplift import utils as uplift_utils
from lightautoml.addons.uplift.metalearners import MetaLearner
from lightautoml.addons.uplift.metalearners import NotTrainedError
from lightautoml.addons.uplift.metalearners import RLearner
from lightautoml.addons.uplift.metalearners import SLearner
from lightautoml.addons.uplift.metalearners import TDLearner
from lightautoml.addons.uplift.metalearners import TLearner
from lightautoml.addons.uplift.metalearners import XLearner
from lightautoml.addons.uplift.metrics import ConstPredictError
from lightautoml.addons.uplift.metrics import TUpliftMetric
from lightautoml.addons.uplift.metrics import calculate_uplift_auc
from lightautoml.automl.base import AutoML
from lightautoml.automl.presets.tabular_presets import TabularAutoML
from lightautoml.report.report_deco import ReportDecoUplift
from lightautoml.tasks import Task
from lightautoml.utils.timer import Timer


logger = logging.getLogger(__name__)


def noner(x, f):
    """None."""
    return None if x is None else f(x)


@dataclass
class Wrapper:
    """Wrapper for class."""

    name: str
    klass: Any
    params: Dict[str, Any]

    def __call__(self):
        """Initiate klass with params."""

        def _init(x: Any):
            if issubclass(x.__class__, Wrapper):
                return x()
            elif isinstance(x, list):
                return [_init(elem) for elem in x]
            elif isinstance(x, dict):
                return {k: _init(v) for k, v in x.items()}
            else:
                return x

        init_params = {k: deepcopy(_init(v)) for k, v in self.params.items()}

        return self.klass(**init_params)

    def update_params(self, d: Dict[str, Any]):
        """Update parameters.

        Rewrite old value of key by new value.

        Args:
            d: Updates.

        """
        self.params.update(d)


class BaseLearnerWrapper(Wrapper):
    """Wrapper for baselearner."""

    pass


class MetaLearnerWrapper(Wrapper):
    """Wrapper for metalearner."""

    def update_baselearner_params(self, d: Dict[str, Any]):
        """Update params."""
        new_params: Dict[str, Any] = {}
        for k, v in self.params.items():
            if isinstance(v, BaseLearnerWrapper):
                v_t = deepcopy(v)
                v_t.update_params(d)
                new_params[k] = v_t
            elif isinstance(v, List):
                vs_t = []
                for x in v:
                    if isinstance(x, BaseLearnerWrapper):
                        x.update_params(d)
                    vs_t.append(x)
                new_params[k] = vs_t
            else:
                new_params[k] = v

        self.params = new_params


class BaseAutoUplift(metaclass=abc.ABCMeta):
    """Base class for AutoUplift.

    Args:
        base_task: Task ('binary'/'reg') if there aren't candidates.
        metric: Uplift metric.
        has_report: Use metalearner for which report can be constructed.
        increasing_metric: Increasing metric
        test_size: Size of test part, which use for.
        timeout: Global timeout of autouplift. Doesn't work when uplift_candidates is not default.
        timeout_metalearner: Timeout for metalearner.
        timeout_single_learner: Timeout single baselearner, if not specified, it's selected automatically.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        random_state: Random state.

    """

    def __init__(
        self,
        base_task: Task,
        metric: Union[str, TUpliftMetric, Callable] = "adj_qini",
        has_report: bool = False,
        increasing_metric: bool = True,
        test_size: float = 0.2,
        timeout: Optional[int] = None,
        timeout_metalearner: Optional[int] = None,
        timeout_single_learner: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        random_state: int = 42,
    ):
        assert 0.0 < test_size < 1.0, "'test_size' must be between (0.0, 1.0)"
        assert (
            isinstance(metric, str) or isinstance(metric, TUpliftMetric) or callable(metric)
        ), "Invalid metric type: available = {'str', 'TUpliftMetric', 'function'}"

        self.base_task = base_task
        self.metric = metric
        self.has_report = has_report
        self.increasing_metric = increasing_metric
        self.test_size = test_size
        self.timeout = timeout
        self.timeout_metalearner = timeout_metalearner
        self.timeout_single_learner = timeout_single_learner
        self.cpu_limit = cpu_limit
        self.gpu_ids = gpu_ids
        self.random_state = random_state

        self._timer = Timer()
        if timeout is not None:
            self._timer._timeout = timeout

    @abstractmethod
    def fit(self, data: DataFrame, roles: Dict, verbose: int = 0):
        """Fit AutoUplift.

        Choose best metalearner and fit it.

        Args:
            data: Dataset to train.
            roles: Roles dict with 'treatment' roles.
            verbose: Verbose.

        """
        pass

    @abstractmethod
    def predict(self, data: Any) -> Tuple[np.ndarray, ...]:
        """Predict treatment effects.

        Predict treatment effects using best metalearner

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            ...: None or predictions of base task values on treated(control)-stage

        """
        pass

    def calculate_metric(self, y_true: np.ndarray, uplift_pred: np.ndarray, treatment: np.ndarray) -> float:
        """Calculate uplift metric.

        Args:
            y_true: Target values
            uplift_pred: Prediction of models
            treatment: Treatment column

        Returns:
            value of calculated metric.

        """
        normed = self.base_task._name == "binary"
        if isinstance(self.metric, str):
            try:
                auc = calculate_uplift_auc(
                    y_true,
                    uplift_pred,
                    treatment,
                    self.metric,
                    normed,
                )
            except ConstPredictError as e:
                logger.error(str(e) + "\nMetric set to zero.")
                auc = 0.0

            return auc
        elif isinstance(self.metric, TUpliftMetric) or callable(self.metric):
            return self.metric(y_true, uplift_pred, treatment)
        else:
            raise Exception("Invalid metric type: available = {'str', 'TUpliftMetric'}")

    def _prepare_data(self, data: DataFrame, roles: dict) -> Tuple[DataFrame, DataFrame, np.ndarray, np.ndarray]:
        """Prepare data for training part.

        Args:
            data: Dataset to train.
            roles: Roles dict with 'treatment' roles.

        Returns:
            train_data: Train part of data
            test_data: Test part of data
            test_treatment: Treatment values of test data
            test_target: Target values of test data

        """
        (
            _,
            target_col,
        ) = uplift_utils._get_target_role(roles)
        (
            _,
            treatment_col,
        ) = uplift_utils._get_treatment_role(roles)

        stratify_value = None if self.base_task._name == "reg" else data[[target_col, treatment_col]]

        train_data, test_data = train_test_split(
            data,
            test_size=self.test_size,
            stratify=stratify_value,
            random_state=self.random_state,
            shuffle=True,
        )
        test_treatment = test_data[treatment_col].ravel()
        test_target = test_data[target_col].ravel()

        return (
            train_data,
            test_data,
            test_treatment,
            test_target,
        )


class AutoUplift(BaseAutoUplift):
    """AutoUplift.

    Using greed-search to choose best uplift-approach.

    Args:
        base_task: Task ('binary'/'reg') if there aren't candidates.
        uplift_candidates: List of metalearners with params and custom name.
        add_dd_candidates: Add data depend candidates. Doesn't work when uplift_candidates is not default.
        metric: Uplift metric.
        has_report: Use metalearner for which report can be constructed.
        increasing_metric: Increasing metric
        test_size: Size of test part, which use for.
        threshold_imbalance_treatment: Threshold for imbalance treatment.
            Condition: | MEAN(treatment) - 0.5| > threshold_imbalance_treatment
        timeout: Global timeout of autouplift. Doesn't work when uplift_candidates is not default.
        timeout_metalearner: Timeout for metalearner.
        timeout_single_learner: Timeout single baselearner, if not specified, it's selected automatically.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        random_state: Random state.


    Attributes:
        _tabular_timeout: Timeout for base learner in Tabularperset
        __THRESHOLD_DISBALANCE_TREATMENT__: Threshold for imbalance treatment.
            Condition: | treatment.mean() - 0.5| > __THRESHOLD_DISBALANCE_TREATMENT__

    """

    def __init__(
        self,
        base_task: Task,
        uplift_candidates: List[MetaLearnerWrapper] = [],
        add_dd_candidates: bool = False,
        metric: Union[str, TUpliftMetric, Callable] = "adj_qini",
        has_report: bool = False,
        increasing_metric: bool = True,
        test_size: float = 0.2,
        threshold_imbalance_treatment: float = 0.2,
        timeout: Optional[int] = None,
        timeout_metalearner: Optional[int] = None,
        timeout_single_learner: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        random_state: int = 42,
    ):
        super().__init__(
            base_task,
            metric,
            has_report,
            increasing_metric,
            test_size,
            timeout,
            timeout_metalearner,
            timeout_single_learner,
            cpu_limit,
            gpu_ids,
            random_state,
        )

        if len(uplift_candidates) > 0:
            if timeout is not None:
                logger.warning("'timeout' is used only for a global time.")
            if add_dd_candidates:
                logger.warning("'add_dd_candidates' isn't used when 'uplift_candidates' is specified.")

        # self.checkout_timeout = True
        if len(uplift_candidates) > 0:
            self.uplift_candidates = uplift_candidates
            # self.checkout_timeout = False
        else:
            self.uplift_candidates = []

        self.best_metalearner: Optional[MetaLearner] = None
        self.best_metalearner_candidate: Optional[MetaLearnerWrapper] = None
        self.add_dd_candidates = add_dd_candidates
        self.candidate_holdout_metrics: List[Union[float, None]] = []
        self._threshold_imbalance_treatment = threshold_imbalance_treatment

    def fit(
        self,
        data: DataFrame = None,
        roles: Dict = None,
        verbose: int = 0,
        train_data: DataFrame = None,
        test_data: DataFrame = None,
    ):
        """Fit AutoUplift.

        Choose best metalearner and fit it.

        Args:
            data: Dataset to train.
            roles: Roles dict with 'treatment' roles.
            verbose: Verbose.
            train_data: Train data.
            test_data: Test data.

        """
        if test_data is None:
            (
                train_data,
                test_data,
                test_treatment,
                test_target,
            ) = self._prepare_data(data, roles)
        else:
            _, target_col = uplift_utils._get_target_role(roles)
            _, treatment_col = uplift_utils._get_treatment_role(roles)
            test_treatment = test_data[treatment_col].ravel()
            test_target = test_data[target_col].ravel()

        best_metalearner: Optional[MetaLearner] = None
        best_metalearner_candidate_info: Optional[MetaLearnerWrapper] = None
        best_metric_value = 0.0

        if len(self.uplift_candidates) == 0:
            self._generate_uplift_candidates(data, roles)

        self.candidate_holdout_metrics = [None] * len(self.uplift_candidates)
        self.candidate_worktime = [None] * len(self.uplift_candidates)

        self._timer.start()

        for (
            idx_candidate,
            candidate_info,
        ) in enumerate(self.uplift_candidates):
            metalearner = candidate_info()

            try:
                start_fit = time.time()
                metalearner.fit(train_data, roles, verbose)
                logger.info(
                    "Uplift candidate #{} [{}] is fitted".format(
                        idx_candidate,
                        candidate_info.name,
                    )
                )
                end_fit = time.time()
                self.candidate_worktime[idx_candidate] = end_fit - start_fit

                (
                    uplift_pred,
                    _,
                    _,
                ) = metalearner.predict(test_data)
                uplift_pred = uplift_pred.ravel()

                metric_value = self.calculate_metric(
                    test_target,
                    uplift_pred,
                    test_treatment,
                )
                self.candidate_holdout_metrics[idx_candidate] = metric_value

                if best_metalearner_candidate_info is None:
                    best_metalearner = metalearner
                    best_metalearner_candidate_info = candidate_info
                    best_metric_value = metric_value
                elif (best_metric_value < metric_value and self.increasing_metric) or (
                    best_metric_value > metric_value and not self.increasing_metric
                ):
                    best_metalearner = metalearner
                    best_metalearner_candidate_info = candidate_info
                    best_metric_value = metric_value
            except NotTrainedError:
                end_fit = time.time()
                self.candidate_worktime[idx_candidate] = end_fit - start_fit

            if self._timer.time_limit_exceeded():
                logger.warning(
                    "Time of training exceeds 'timeout': {} > {}.".format(
                        self._timer.time_spent,
                        self.timeout,
                    )
                )
                logger.warning(
                    "There is fitted {}/{} candidates".format(
                        idx_candidate + 1,
                        len(self.uplift_candidates),
                    )
                )
                if idx_candidate + 1 < len(self.uplift_candidates):
                    logger.warning("Try to increase 'timeout' or set 'None'(eq. infinity)")
                break

        self.best_metalearner_candidate_info = best_metalearner_candidate_info
        self.best_metalearner = best_metalearner

    def predict(self, data: DataFrame) -> Tuple[np.ndarray, ...]:
        """Predict treatment effects.

        Predict treatment effects using best metalearner

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            ...: None or predictions of base task values on treated(control)-stage

        """
        assert self.best_metalearner is not None, "First call 'self.fit(...)', to choose best metalearner"

        return self.best_metalearner.predict(data)

    def create_best_metalearner(
        self,
        need_report: bool = False,
        update_metalearner_params: Dict[str, Any] = {},
        update_baselearner_params: Dict[str, Any] = {},
    ) -> Union[MetaLearner, ReportDecoUplift]:
        """Create 'raw' best metalearner with(without) report functionality.

        Returned metalearner should be refitted.

        Args:
            need_report: Wrap best metalearner into Report
            update_metalearner_params: MetaLearner parameters.
            update_baselearner_params: Parameters inner learner.
                Recommended using - increasing timeout of 'TabularAutoML' learner for better scores.
                Example: {'timeout': None}.

        Returns:
            metalearner_deco: Best metalearner is wrapped or not by ReportDecoUplift.

        """
        assert self.best_metalearner_candidate_info is not None, "First call 'self.fit(...), to choose best metalearner"

        candidate_info = deepcopy(self.best_metalearner_candidate_info)
        if len(update_metalearner_params) > 0:
            candidate_info.update_params(update_metalearner_params)
        if len(update_baselearner_params) > 0:
            candidate_info.update_baselearner_params(update_baselearner_params)

        best_metalearner = candidate_info()

        if need_report:
            if isinstance(self.metric, str):
                rdu = ReportDecoUplift()
                best_metalearner = rdu(best_metalearner)
            else:
                logger.warning("Report doesn't work with custom metric, return just best_metalearner.")

        return best_metalearner

    def get_metalearners_rating(self) -> DataFrame:
        """Get rating of metalearners.

        Returns:
            DataFrame with rating.

        """
        rating_table = DataFrame(
            {
                "MetaLearner": [info.name for info in self.uplift_candidates],
                "Parameters": [info.params for info in self.uplift_candidates],
                "Metrics": self.candidate_holdout_metrics,
                "WorkTime": self.candidate_worktime,
            }
        )

        rating_table["Rank"] = rating_table["Metrics"].rank(
            method="first",
            ascending=not self.increasing_metric,
        )
        rating_table.sort_values("Rank", inplace=True)
        rating_table.reset_index(drop=True, inplace=True)

        return rating_table

    def _generate_uplift_candidates(self, data: DataFrame, roles):
        """Generate uplift candidates.

        Combine uplift candidates from 'default' and 'data-depends' candidates.

        Args:
            data: Dataset to train.
            roles: Roles dict with 'treatment' roles.

        """
        self._calculate_tabular_time()

        self.uplift_candidates = self._default_uplift_candidates
        if self.has_report:
            self.uplift_candidates = [
                c for c in self.uplift_candidates if c.klass in ReportDecoUplift._available_metalearners
            ]

        if self.add_dd_candidates:
            dd_candidates = self._generate_data_depend_uplift_candidates(data, roles)
            if self.has_report:
                dd_candidates = [c for c in dd_candidates if c.klass in ReportDecoUplift._available_metalearners]
            self.uplift_candidates.extend(dd_candidates)

    def _calculate_tabular_time(self):
        # Number TabularAutoML in all possible uplift candidates
        # TODO: rewrite this
        if self.has_report:
            num_tabular_automls = 16 if self.add_dd_candidates else 11
        else:
            num_tabular_automls = 22 if self.add_dd_candidates else 17

        if self.timeout_single_learner is not None:
            self._tabular_timeout = self.timeout_single_learner
        elif self.timeout_metalearner is not None:
            self._tabular_timeout = None
        elif self.timeout is not None:
            self._tabular_timeout = self.timeout / num_tabular_automls
        else:
            self._tabular_timeout = None

    @property
    def _default_uplift_candidates(self) -> List[MetaLearnerWrapper]:
        """Default uplift candidates."""
        return [
            MetaLearnerWrapper(
                name="__SLearner__Default__",
                klass=SLearner,
                params={
                    "base_task": self.base_task,
                    "timeout": self.timeout_metalearner,
                },
            ),
            MetaLearnerWrapper(
                name="__TLearner__Default__",
                klass=TLearner,
                params={
                    "base_task": self.base_task,
                    "timeout": self.timeout_metalearner,
                },
            ),
            MetaLearnerWrapper(
                name="__TDLearner__Default__",
                klass=TDLearner,
                params={
                    "base_task": self.base_task,
                    "timeout": self.timeout_metalearner,
                },
            ),
            MetaLearnerWrapper(
                name="__XLearner__Default__",
                klass=XLearner,
                params={
                    "base_task": self.base_task,
                    "timeout": self.timeout_metalearner,
                },
            ),
            MetaLearnerWrapper(
                name="__RLearner__Linear__",
                klass=RLearner,
                params={
                    "timeout": self.timeout_metalearner,
                    "propensity_learner": BaseLearnerWrapper(
                        name="__Linear__",
                        klass=uplift_utils.create_linear_automl,
                        params={"task": Task("binary")},
                    ),
                    "mean_outcome_learner": BaseLearnerWrapper(
                        name="__Linear__",
                        klass=uplift_utils.create_linear_automl,
                        params={"task": self.base_task},
                    ),
                    "effect_learner": BaseLearnerWrapper(
                        name="__Linear__",
                        klass=uplift_utils.create_linear_automl,
                        params={"task": Task("reg")},
                    ),
                },
            ),
            MetaLearnerWrapper(
                name="__RLearner__Default__",
                klass=RLearner,
                params={
                    "base_task": self.base_task,
                    "timeout": self.timeout_metalearner
                    if self.timeout_metalearner is not None
                    else (self._tabular_timeout * 3 if self._tabular_timeout is not None else None),
                },
            ),
            MetaLearnerWrapper(
                name="__SLearner__TabularAutoML__",
                klass=SLearner,
                params={
                    "timeout": self.timeout_metalearner,
                    "learner": BaseLearnerWrapper(
                        name="__TabularAutoML__",
                        klass=TabularAutoML,
                        params={
                            "task": self.base_task,
                            "timeout": self.timeout_metalearner
                            if self._tabular_timeout is None
                            else self._tabular_timeout,
                        },
                    ),
                },
            ),
            MetaLearnerWrapper(
                name="__TLearner__TabularAutoML__",
                klass=TLearner,
                params={
                    "timeout": self.timeout_metalearner,
                    "treatment_learner": BaseLearnerWrapper(
                        name="__TabularAutoML__",
                        klass=TabularAutoML,
                        params={
                            "task": self.base_task,
                            "timeout": noner(self.timeout_metalearner, lambda t: int(t / 2))
                            if self._tabular_timeout is None
                            else self._tabular_timeout,
                        },
                    ),
                    "control_learner": BaseLearnerWrapper(
                        name="__TabularAutoML__",
                        klass=TabularAutoML,
                        params={
                            "task": self.base_task,
                            "timeout": noner(self.timeout_metalearner, lambda t: int(t / 2))
                            if self._tabular_timeout is None
                            else self._tabular_timeout,
                        },
                    ),
                },
            ),
            MetaLearnerWrapper(
                name="__TDLearner__TabularAutoML__",
                klass=TDLearner,
                params={
                    "timeout": self.timeout_metalearner,
                    "treatment_learner": BaseLearnerWrapper(
                        name="__TabularAutoML__",
                        klass=TabularAutoML,
                        params={
                            "task": self.base_task,
                            "timeout": noner(self.timeout_metalearner, lambda t: int(t / 2))
                            if self._tabular_timeout is None
                            else self._tabular_timeout,
                        },
                    ),
                    "control_learner": BaseLearnerWrapper(
                        name="__TabularAutoML__",
                        klass=TabularAutoML,
                        params={
                            "task": self.base_task,
                            "timeout": noner(self.timeout_metalearner, lambda t: int(t / 2))
                            if self._tabular_timeout is None
                            else self._tabular_timeout,
                        },
                    ),
                },
            ),
            MetaLearnerWrapper(
                name="__XLearner__Propensity_Linear__Other_TabularAutoML__",
                klass=XLearner,
                params={
                    "timeout": self.timeout_metalearner,
                    "outcome_learners": [
                        BaseLearnerWrapper(
                            name="__TabularAutoML__",
                            klass=TabularAutoML,
                            params={
                                "task": self.base_task,
                                "timeout": noner(self.timeout_metalearner, lambda t: int(t / 4))
                                if self._tabular_timeout is None
                                else self._tabular_timeout,
                            },
                        )
                    ],
                    "effect_learners": [
                        BaseLearnerWrapper(
                            name="__TabularAutoML__",
                            klass=TabularAutoML,
                            params={
                                "task": Task("reg"),
                                "timeout": noner(self.timeout_metalearner, lambda t: int(t / 4))
                                if self._tabular_timeout is None
                                else self._tabular_timeout,
                            },
                        )
                    ],
                    "propensity_learner": BaseLearnerWrapper(
                        name="__Linear__",
                        klass=uplift_utils.create_linear_automl,
                        params={"task": Task("binary")},
                    ),
                },
            ),
            MetaLearnerWrapper(
                name="__XLearner__TabularAutoML__",
                klass=XLearner,
                params={
                    "timeout": self.timeout_metalearner,
                    "outcome_learners": [
                        BaseLearnerWrapper(
                            name="__TabularAutoML__",
                            klass=TabularAutoML,
                            params={
                                "task": self.base_task,
                                "timeout": noner(self.timeout_metalearner, lambda t: int(t / 5))
                                if self._tabular_timeout is None
                                else self._tabular_timeout,
                            },
                        )
                    ],
                    "effect_learners": [
                        BaseLearnerWrapper(
                            name="__TabularAutoML__",
                            klass=TabularAutoML,
                            params={
                                "task": Task("reg"),
                                "timeout": noner(self.timeout_metalearner, lambda t: int(t / 5))
                                if self._tabular_timeout is None
                                else self._tabular_timeout,
                            },
                        )
                    ],
                    "propensity_learner": BaseLearnerWrapper(
                        name="__TabularAutoML__",
                        klass=TabularAutoML,
                        params={
                            "task": Task("binary"),
                            "timeout": noner(self.timeout_metalearner, lambda t: int(t / 5))
                            if self._tabular_timeout is None
                            else self._tabular_timeout,
                        },
                    ),
                },
            ),
        ]

    def _generate_data_depend_uplift_candidates(self, data: DataFrame, roles: dict) -> List[MetaLearnerWrapper]:
        """Generate uplift candidates.

        Generate new uplift candidates which depend from data.

        If there is imbalance in treatment , adds the simple linear model for smaller stage.

        Args:
            data: Dataset to train.
            roles: Roles dict with 'treatment' roles.

        Returns:
            List of new uplift candidates.

        """
        dd_uplift_candidates: List[MetaLearnerWrapper] = []

        _, treatment_col = uplift_utils._get_treatment_role(roles)

        treatment_rate = data[treatment_col].mean()

        is_imbalance_treatment = False

        ordered_outcome_learners = [
            BaseLearnerWrapper(
                name="__Linear__",
                klass=uplift_utils.create_linear_automl,
                params={"task": Task("binary")},
            ),
            BaseLearnerWrapper(
                name="__TabularAutoML__",
                klass=TabularAutoML,
                params={
                    "task": self.base_task,
                    "timeout": self._tabular_timeout,
                },
            ),
        ]
        ordered_effect_learners = [
            BaseLearnerWrapper(
                name="__Linear__",
                klass=uplift_utils.create_linear_automl,
                params={"task": Task("reg")},
            ),
            BaseLearnerWrapper(
                name="__TabularAutoML__",
                klass=TabularAutoML,
                params={
                    "task": Task("reg"),
                    "timeout": self._tabular_timeout,
                },
            ),
        ]
        control_model, treatment_model = (
            "Linear",
            "Preset",
        )

        if treatment_rate > 0.5 + self._threshold_imbalance_treatment:
            is_imbalance_treatment = True
        elif treatment_rate < 0.5 - self._threshold_imbalance_treatment:
            is_imbalance_treatment = True
            ordered_outcome_learners = ordered_outcome_learners[::-1]
            ordered_effect_learners = ordered_effect_learners[::-1]
            control_model, treatment_model = "Preset", "Linear"

        if is_imbalance_treatment:
            dd_uplift_candidates.extend(
                [
                    MetaLearnerWrapper(
                        name="XLearner__Propensity_Linear__Control_{}__Treatment_{}".format(
                            control_model,
                            treatment_model,
                        ),
                        klass=XLearner,
                        params={
                            "timeout": self.timeout_metalearner,
                            "outcome_learners": ordered_outcome_learners,
                            "effect_learners": ordered_effect_learners,
                            "propensity_learner": BaseLearnerWrapper(
                                name="__Linear__",
                                klass=uplift_utils.create_linear_automl,
                                params={"task": Task("binary")},
                            ),
                        },
                    ),
                    MetaLearnerWrapper(
                        name="XLearner__Control_{}__Treatment_{}".format(
                            control_model,
                            treatment_model,
                        ),
                        klass=XLearner,
                        params={
                            "timeout": self.timeout_metalearner,
                            "outcome_learners": ordered_outcome_learners,
                            "effect_learners": ordered_effect_learners,
                            "propensity_learner": BaseLearnerWrapper(
                                name="__TabularAutoML__",
                                klass=TabularAutoML,
                                params={
                                    "task": Task("binary"),
                                    "timeout": self._tabular_timeout,
                                },
                            ),
                        },
                    ),
                ]
            )

        return dd_uplift_candidates


MLStageFullName = Tuple[str, ...]
TrainedMetaLearnerFullName = Tuple[str, Tuple[Tuple[MLStageFullName, str], ...]]


@dataclass
class MetaLearnerStage:
    """Metalearner stage."""

    name: str
    params: Dict[str, Any] = field(default_factory=dict)
    prev_stage: Optional["MetaLearnerStage"] = None

    def full_name(self) -> MLStageFullName:
        """Get full name."""
        fn: MLStageFullName
        if self.prev_stage is None:
            fn = (self.name,)
        else:
            fn = (
                *self.prev_stage.full_name(),
                self.name,
            )

        return fn


@dataclass
class TrainedStageBaseLearner:
    """Trained stage baselearner."""

    stage_bl: BaseLearnerWrapper
    prev_stage_bl: Optional[BaseLearnerWrapper]
    trained_model: AutoML
    pred: np.ndarray


class AutoUpliftTX(BaseAutoUplift):
    """AutoUplift for T(X)-Learners.

    Optimizes the selection of best metalearner between TLearner and XLearner,
    without don't retrain the baselearners of the common parts. TLearner is the first half of XLearner.

    Args:
        base_task: Task ('binary'/'reg') if there aren't candidates.
        baselearners: List of baselearners or baselearner divided into the groups (Dict).
        metalearners: List of metalearners.
        metric: Uplift metric.
        increasing_metric: Increasing metric.
        test_size: Size of test part, which use for.
        timeout: Global timeout of autouplift. Doesn't work when uplift_candidates is not default.
        timeout_single_learner: Timeout single baselearner, if not specified, it's selected automatically.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        random_state: Random state.

    """

    __MAP_META_TO_STAGES__: Dict[str, List[MetaLearnerStage]] = {
        "TLearner": [
            MetaLearnerStage(name="outcome_control"),
            MetaLearnerStage(name="outcome_treatment"),
        ],
        "XLearner": [
            MetaLearnerStage(name="outcome_control"),
            MetaLearnerStage(name="outcome_treatment"),
            MetaLearnerStage(
                name="propensity",
                params={"task": Task("binary")},
            ),
            MetaLearnerStage(
                name="effect_control",
                params={"task": Task("reg")},
                prev_stage=MetaLearnerStage(name="outcome_treatment"),
            ),
            MetaLearnerStage(
                name="effect_treatment",
                params={"task": Task("reg")},
                prev_stage=MetaLearnerStage(name="outcome_control"),
            ),
        ],
    }

    def __init__(
        self,
        base_task: Task,
        baselearners: Optional[
            Union[
                List[BaseLearnerWrapper],
                Dict[
                    MLStageFullName,
                    List[BaseLearnerWrapper],
                ],
            ]
        ] = None,
        metalearners: List[str] = [],
        metric: Union[str, TUpliftMetric, Callable] = "adj_qini",
        increasing_metric: bool = True,
        test_size: float = 0.2,
        timeout: Optional[int] = None,
        timeout_single_learner: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        random_state: int = 42,
    ):
        assert all(ml in self.__MAP_META_TO_STAGES__ for ml in metalearners), "Currently available for {}.".format(
            self.__MAP_META_TO_STAGES__
        )

        super().__init__(
            base_task,
            metric,
            True,
            increasing_metric,
            test_size,
            timeout,
            None,
            timeout_single_learner,
            cpu_limit,
            gpu_ids,
            random_state,
        )

        self.baselearners = baselearners
        self.metalearners = metalearners if len(metalearners) > 0 else list(self.__MAP_META_TO_STAGES__)

        self._best_metalearner: MetaLearner
        self._best_metalearner_wrap: MetaLearnerWrapper

        self._trained_stage_baselearners: Dict[
            MLStageFullName,
            List[TrainedStageBaseLearner],
        ] = defaultdict(list)
        self._metalearner_metrics: Dict[TrainedMetaLearnerFullName, float] = {}

        self._n_run_l2 = 3

    def fit(self, data: DataFrame, roles: dict, verbose: int = 0):
        """Fit AutoUplift.

        Choose best metalearner and fit it.

        Args:
            data: Dataset to train.
            roles: Roles dict with 'treatment' roles.
            verbose: Verbose.

        """
        train_data, test_data, test_treatment, test_target = self._prepare_data(data, roles)

        self._timer.start()

        for stage_info in self._generate_stage_baselearner_candidates():
            self._evaluate(stage_info, train_data, test_data, roles, verbose)

            if self._timer.time_limit_exceeded():
                logger.warning(
                    "Time of training exceeds 'timeout': {} > {}.".format(
                        self._timer.time_spent,
                        self.timeout,
                    )
                )
                break

        self._calculate_metalearners_metrics(test_treatment, test_target)

        self._set_best_metalearner()

    def predict(self, data: DataFrame) -> Tuple[np.ndarray, ...]:
        """Predict treatment effects.

        Predict treatment effects using best metalearner

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            ...: None or predictions of base task values on treated(control)-group

        """
        assert self._best_metalearner is not None, "First call 'self.fit(...)', to choose best metalearner."

        return self._best_metalearner.predict(data)

    def create_best_metalearner(
        self,
        need_report: bool = True,
        update_metalearner_params: Dict[str, Any] = {},
        update_baselearner_params: Dict[str, Any] = {},
    ) -> Union[MetaLearner, ReportDecoUplift]:
        """Create 'raw' best metalearner with(without) report functionality.

        Returned metalearner should be refitted.

        Args:
            need_report: Wrap best metalearner into Report
            update_metalearner_params: MetaLearner parameters.
            update_baselearner_params: Parameters inner learner.
                Recommended using - increasing timeout of 'TabularAutoML' learner for better scores.
                Example: {'timeout': None}.

        Returns:
            metalearner_deco: Best metalearner is wrapped or not by ReportDecoUplift.

        """
        assert len(self._trained_stage_baselearners) > 0, "First call 'self.fit(...), to choose best metalearner."

        ml_wrap = deepcopy(self._best_metalearner_wrap)

        if len(update_metalearner_params) > 0:
            ml_wrap.update_params(update_metalearner_params)
        if len(update_baselearner_params) > 0:
            ml_wrap.update_baselearner_params(update_baselearner_params)

        best_metalearner_raw = ml_wrap()

        if need_report:
            if isinstance(self.metric, str):
                rdu = ReportDecoUplift()
                best_metalearner_raw = rdu(best_metalearner_raw)
            else:
                logger.warning("Report doesn't work with custom metric, return just best_metalearner.")

        return best_metalearner_raw

    def get_metalearners_rating(self) -> DataFrame:
        """Get rating of metalearners.

        Returns:
            DataFrame with rating.

        """
        metalearner_names, params, metrics = [], [], []

        for ml_name, metric in self._metalearner_metrics.items():
            metalearner_names.append(ml_name[0])
            params.append(ml_name[1])
            metrics.append(metric)

        rating_table = DataFrame(
            {
                "MetaLearner": metalearner_names,
                "Parameters": params,
                "Metrics": metrics,
            }
        )

        rating_table["Rank"] = rating_table["Metrics"].rank(method="first", ascending=not self.increasing_metric)
        rating_table.sort_values("Rank", inplace=True)
        rating_table.reset_index(drop=True, inplace=True)

        return rating_table

    def _generate_stage_baselearner_candidates(
        self,
    ) -> Generator[Tuple[Tuple[MetaLearnerStage, BaseLearnerWrapper], ...], None, None]:
        """Iterate through a stage of baselearners one at a time."""
        stage_baselearners = self._set_stage_baselearners()

        stage_by_levels = defaultdict(list)
        for stage in stage_baselearners:
            stage_by_levels[len(stage)].append(stage)

        pool_iter_levels = defaultdict(list)

        # TODO - WARNING! Work only with two levels.
        bls_level_1 = zip_longest(
            *(deepcopy(bls) for stage, bls in stage_baselearners.items() if stage in stage_by_levels[1])
        )
        first_run = True
        n_runs = 0
        for bls in bls_level_1:
            for stage_name_l1, bl_l1 in zip(stage_by_levels[1], bls):
                stage_l1 = self._extract_stage(stage_name_l1)

                for stage_name_l2 in stage_by_levels[2]:
                    if stage_name_l2[0:1] == stage_name_l1:
                        pool_iter_levels[2].append(
                            (
                                stage_name_l2,
                                bl_l1,
                                iter(deepcopy(stage_baselearners[stage_name_l2])),
                            )
                        )

                n_runs += 1
                yield ((stage_l1, bl_l1),)

                if not first_run:
                    n_stage_iters = len(pool_iter_levels[2])
                    if n_stage_iters == 0:
                        continue

                    for _ in range(min(n_stage_iters, self._n_run_l2)):
                        n_stage_iters = len(pool_iter_levels[2])
                        idx = np.random.randint(0, n_stage_iters, 1)[0]
                        try:
                            (stage_name_l2, bl_l1, bls_iter,) = pool_iter_levels[
                                2
                            ][idx]
                            bl_l2 = next(bls_iter)

                            stage_l1 = self._extract_stage(stage_name_l2[0:1])
                            stage_l2 = self._extract_stage(stage_name_l2)

                            n_runs += 1
                            yield ((stage_l1, bl_l1), (stage_l2, bl_l2))

                        except StopIteration:
                            pool_iter_levels[2].pop(idx)

                if n_runs >= len(stage_by_levels[1]):
                    first_run = False

        while len(pool_iter_levels[2]) > 0:
            n_stage_iters = len(pool_iter_levels[2])
            try:
                idx = np.random.randint(0, n_stage_iters, 1)[0]

                stage_name_l2, bl_l1, bls_iter = pool_iter_levels[2][idx]
                bl_l2 = next(bls_iter)

                stage_l1 = self._extract_stage(stage_name_l2[0:1])
                stage_l2 = self._extract_stage(stage_name_l2)

                yield ((stage_l1, bl_l1), (stage_l2, bl_l2))
            except StopIteration:
                pool_iter_levels[2].pop(idx)

    def _extract_stages(
        self,
    ) -> Generator[Tuple[str, MetaLearnerStage], None, None]:
        """Iterate over stages."""
        for (
            ml_name,
            ml_stages,
        ) in self.__MAP_META_TO_STAGES__.items():
            for stage in ml_stages:
                yield ml_name, stage

    def _extract_stage(self, full_name: MLStageFullName) -> MetaLearnerStage:
        """Return the first stage with a specific name."""
        for (
            ml_name,
            stage,
        ) in self._extract_stages():
            if stage.full_name() == full_name:
                return stage
        raise Exception("Can't find stage {}".format(full_name))

    def _set_stage_baselearners(
        self,
    ) -> Dict[MLStageFullName, List[BaseLearnerWrapper]]:
        """Generate baselearner for metalearners' stages."""
        # TODO Timeout!
        # baselearners = []
        # if isinstance(self.baselearners, list):
        #     baselearners = deepcopy(self._validate_baselearners(self.baselearners))

        stage_baselearners = {}
        if isinstance(self.baselearners, dict):
            stage_baselearners = {k: deepcopy(self._validate_baselearners(v)) for k, v in self.baselearners.items()}

        all_stages_full_names = set(
            stage.full_name() for ml, ml_stages in self.__MAP_META_TO_STAGES__.items() for stage in ml_stages
        )

        if len(stage_baselearners) != len(all_stages_full_names):
            timeout = self._calculate_single_bl_timeout(len(stage_baselearners) > 0)

            if isinstance(self.baselearners, list) and len(self.baselearners) > 0:
                baselearners = deepcopy(self._validate_baselearners(self.baselearners))
            elif self.baselearners is None or isinstance(self.baselearners, dict):
                baselearners = deepcopy(self.__default_learners(tab_params={"timeout": timeout}))

            remain_stages_full_names = all_stages_full_names - set(stage_baselearners)

            # TODO: Consider the parameters of stages, not only names.
            bin_baselearners: List[BaseLearnerWrapper] = []
            reg_baselearners: List[BaseLearnerWrapper] = []
            raw_baselearners: List[BaseLearnerWrapper] = []

            for bl in baselearners:
                if "task" in bl.params:
                    if bl.params["task"].name == "binary":
                        bin_baselearners.append(bl)
                    if bl.params["task"].name == "reg":
                        reg_baselearners.append(bl)
                else:
                    raw_baselearners.append(bl)

            for full_name in remain_stages_full_names:
                stage = self._extract_stage(full_name)

                stage_task = stage.params["task"] if "task" in stage.params else self.base_task

                filled_baselearners = deepcopy(raw_baselearners)
                for idx in range(len(filled_baselearners)):
                    filled_baselearners[idx].params["task"] = stage_task

                baselearners_on_stage = filled_baselearners
                if stage_task.name == "binary":
                    baselearners_on_stage.extend(bin_baselearners)
                elif stage_task.name == "reg":
                    baselearners_on_stage.extend(reg_baselearners)

                stage_baselearners[full_name] = baselearners_on_stage

        return stage_baselearners

    def _calculate_single_bl_timeout(self, specify_stages: bool) -> Optional[int]:
        """Calculate timeout for single TabularAutoML from default baselearners."""
        timeout: Optional[int] = None
        if not specify_stages:
            if self.timeout is not None:
                timeout = int(
                    self.timeout / (2 * ("TLearner" in self.metalearners) + 5 * ("XLearner" in self.metalearners))
                )
            elif self.timeout_single_learner is not None:
                timeout = self.timeout_single_learner
        else:
            timeout = self.timeout_single_learner

        return timeout

    def _validate_baselearners(self, baselearners: List[BaseLearnerWrapper]):
        """Validate baselearners names.

        Validates baselearners by unique names for equal bl class names.

        Args:
            baselearners: List of baselearners.

        Returns:
            baselearners: Valideted baselearners.

        """
        k2n: Dict[str, List[int]] = {}
        for idx, bl in enumerate(baselearners):
            bl_name = bl.name

            k2n.setdefault(bl_name, [])
            k2n[bl_name].append(idx)

        is_name_duplicates = any(len(idxs) > 1 for bl_name, idxs in k2n.items())

        if is_name_duplicates:
            logger.warning("Naming of baselearner should be unique.")
            logger.warning("Name of baselearner would be updated with postfix '__order_idx_bl__'.")

            renaming_by_idxs: Dict[int, str] = {}
            for bl_name, idxs in k2n.items():
                if len(idxs) > 1:
                    for idx in idxs:
                        renaming_by_idxs[idx] = "{}__#{}__".format(bl_name, idx)

            baselearners_t = []
            for idx, bl in enumerate(baselearners):
                if idx in renaming_by_idxs:
                    bl.name = renaming_by_idxs[idx]

                baselearners_t.append(bl)
            return baselearners_t
        else:
            return baselearners

    def _evaluate(
        self,
        stage_info: Tuple[
            Tuple[
                MetaLearnerStage,
                BaseLearnerWrapper,
            ],
            ...,
        ],
        train: DataFrame,
        test: DataFrame,
        roles: dict,
        verbose: int = 0,
    ):
        """Evaluate baselearner: fit-train/predict-test."""
        train_data, train_roles = self._prepare_data_for_stage(stage_info, train, roles)

        ml_stage, bl_wrap = stage_info[-1]
        prev_stage, prev_bl_wrap = None, None
        if len(stage_info) == 2:
            prev_stage, prev_bl_wrap = stage_info[0]

        bl = bl_wrap()

        bl.fit_predict(train_data, train_roles, verbose)
        test_pred = bl.predict(test).data.ravel()

        tsbl = TrainedStageBaseLearner(
            stage_bl=bl_wrap,
            prev_stage_bl=prev_bl_wrap,
            trained_model=bl,
            pred=test_pred,
        )
        self._trained_stage_baselearners[ml_stage.full_name()].append(tsbl)

    def _prepare_data_for_stage(
        self,
        stage_info: Tuple[
            Tuple[
                MetaLearnerStage,
                BaseLearnerWrapper,
            ],
            ...,
        ],
        train: DataFrame,
        roles: dict,
    ) -> Tuple[DataFrame, Dict]:
        """Prepare data and roles for one metalearner's stage training."""
        treatment_role, treatment_col = uplift_utils._get_treatment_role(roles)
        target_role, target_col = uplift_utils._get_target_role(roles)

        stage_name = stage_info[-1][0].name

        if len(stage_info) == 1:
            if stage_name == "propensity":
                train_roles = deepcopy(roles)
                train_roles.pop(treatment_role)
                train_roles.pop(target_role)
                train_roles["target"] = treatment_col
                train_data = train.drop(target_col, axis=1)
            elif stage_name == "outcome_control":
                train_roles = deepcopy(roles)
                train_roles.pop(treatment_role)
                train_data = train[train[treatment_col] == 0].drop(treatment_col, axis=1)
            elif stage_name == "outcome_treatment":
                train_roles = deepcopy(roles)
                train_roles.pop(treatment_role)
                train_data = train[train[treatment_col] == 1].drop(treatment_col, axis=1)
            else:
                raise Exception("Wrong l1 stage name")
        elif len(stage_info) == 2:
            train_roles = deepcopy(roles)
            train_roles.pop(treatment_role)

            prev_ml_stage, prev_bl_wrap = stage_info[0]

            prev_stage_bl = [
                bl.trained_model
                for bl in self._trained_stage_baselearners[prev_ml_stage.full_name()]
                if bl.stage_bl.name == prev_bl_wrap.name
            ][0]

            if stage_name == "effect_control":
                train_data = train[train[treatment_col] == 0].drop(treatment_col, axis=1)
                opposite_gr_pred = prev_stage_bl.predict(train_data).data.ravel()

                train_data[target_col] = opposite_gr_pred - train_data[target_col]
            elif stage_name == "effect_treatment":
                train_data = train[train[treatment_col] == 1].drop(treatment_col, axis=1)
                opposite_gr_pred = prev_stage_bl.predict(train_data).data.ravel()

                train_data[target_col] = train_data[target_col] - opposite_gr_pred
            else:
                raise Exception("Wrong l2 stage name")

        return train_data, train_roles

    def _calculate_metalearners_metrics(
        self,
        test_target: np.ndarray,
        test_treatment: np.ndarray,
    ):
        """Calculate metalearners' metric."""
        for set_ml_stage_bls in self._bl_for_ml():
            for (
                ml_name,
                stage_bls,
            ) in set_ml_stage_bls.items():
                sbls = tuple(
                    sorted(
                        [
                            (
                                stage_name,
                                bl.stage_bl.name,
                            )
                            for stage_name, bl in stage_bls.items()
                        ]
                    )
                )

                trained_ml_full_name = (
                    ml_name,
                    sbls,
                )

                uplift_pred = self._metalearner_predict(ml_name, stage_bls)
                metric_value = self.calculate_metric(
                    test_target,
                    uplift_pred,
                    test_treatment,
                )
                self._metalearner_metrics[trained_ml_full_name] = metric_value

    def _bl_for_ml(
        self,
    ) -> Generator[Dict[str, Dict[MLStageFullName, TrainedStageBaseLearner]], None, None]:
        """Prepare stage-baselearners for calculating metalearner metrics."""
        ready_metalearners = []
        for (
            ml_name,
            ml_stages,
        ) in self.__MAP_META_TO_STAGES__.items():
            ready_metalearners.append(all(s.full_name() in self._trained_stage_baselearners for s in ml_stages))

        if not any(ready_metalearners):
            raise Exception("No one metalearner can predict.")

        stage_baselearners = {}
        for (
            stage_fullname,
            bls,
        ) in self._trained_stage_baselearners.items():
            stage_baselearners[stage_fullname] = bls

        stage_names = list(stage_baselearners.keys())
        bls_prd = product(*(bls for _, bls in stage_baselearners.items()))

        for bl in bls_prd:
            set_bls = dict(zip(stage_names, bl))

            set_ml_with_sbls = {}
            for (
                ml_name,
                ml_stages,
            ) in self.__MAP_META_TO_STAGES__.items():
                ml_bls = {}
                for ml_stage in ml_stages:
                    ml_stage_full_name = ml_stage.full_name()
                    if ml_stage_full_name in set_bls:
                        trained_sbl = set_bls[ml_stage_full_name]
                        if trained_sbl.prev_stage_bl is None:
                            ml_bls[ml_stage_full_name] = trained_sbl
                        else:
                            if not ml_stage_full_name[0:1] in set_bls:
                                continue

                            if trained_sbl.prev_stage_bl.name == set_bls[ml_stage_full_name[0:1]].stage_bl.name:
                                ml_bls[ml_stage_full_name] = set_bls[ml_stage_full_name]

                if len(ml_bls) == len(ml_stages):
                    set_ml_with_sbls[ml_name] = ml_bls

            yield set_ml_with_sbls

    def _metalearner_predict(
        self, metalearner: str, baselearners: Dict[MLStageFullName, TrainedStageBaseLearner]
    ) -> np.ndarray:
        """Metalearners prediction.

        Use pre-calculated scores of baselearners.

        Args:
            metalearner: Metalearner name.
            baselearners: Mapping metalearner stage name to trained baselearner.

        Returns:
            uplift_pred: Prediction.

        """
        if metalearner == "TLearner":
            control_pred = baselearners[("outcome_control",)].pred
            treatment_pred = baselearners[("outcome_treatment",)].pred
            uplift_pred = treatment_pred - control_pred
        elif metalearner == "XLearner":
            control_pred = baselearners[
                (
                    "outcome_treatment",
                    "effect_control",
                )
            ].pred
            treatment_pred = baselearners[
                (
                    "outcome_control",
                    "effect_treatment",
                )
            ].pred
            propensity_pred = baselearners[("propensity",)].pred
            uplift_pred = propensity_pred * treatment_pred + (1 - propensity_pred) * control_pred
        else:
            raise Exception()

        return uplift_pred.ravel()

    def _set_best_metalearner(self):
        """Select the best metalearner from the trained ones."""
        best_metric_value = None
        best_candidate = None
        for k, v in self._metalearner_metrics.items():
            if best_metric_value is None:
                best_candidate = k
                best_metric_value = v
            elif (self.increasing_metric and best_metric_value < v) or (
                not self.increasing_metric and best_metric_value > v
            ):
                best_candidate = k
                best_metric_value = v

        ml_name, stages_params = best_candidate
        stages_params = dict(stages_params)

        self._best_metalearner = self._init_metalearner(ml_name, stages_params)
        self._best_metalearner_wrap = self._create_metalearner_wrap(ml_name, stages_params)

    def _init_metalearner(self, metalearner_name: str, bls: Dict[MLStageFullName, str]) -> MetaLearner:
        """Initialize best metalearner from trained baselearners.

        Args:
            metalearner_name: Metalearner name.
            bls: Mapping metalearner stage to baselearner name.

        Returns:
            Metalearner.

        """
        ml: Optional[MetaLearner] = None
        if metalearner_name == "TLearner":
            ocl = self._get_trained_bl(("outcome_control",), bls[("outcome_control",)]).trained_model
            otl = self._get_trained_bl(("outcome_treatment",), bls[("outcome_treatment",)]).trained_model

            ml = TLearner(control_learner=ocl, treatment_learner=otl)
        elif metalearner_name == "XLearner":
            ocl = self._get_trained_bl(("outcome_control",), bls[("outcome_control",)]).trained_model
            otl = self._get_trained_bl(("outcome_treatment",), bls[("outcome_treatment",)]).trained_model
            pl = self._get_trained_bl(("propensity",), bls[("propensity",)]).trained_model
            ecl = self._get_trained_bl(
                ("outcome_treatment", "effect_control"), bls[("outcome_treatment", "effect_control")]
            ).trained_model
            etl = self._get_trained_bl(
                ("outcome_control", "effect_treatment"),
                bls[("outcome_control", "effect_treatment")],
            ).trained_model

            ml = XLearner(outcome_learners=[ocl, otl], effect_learners=[ecl, etl], propensity_learner=pl)
        else:
            raise Exception()

        return ml

    def _get_trained_bl(
        self,
        metalearner_stage: MLStageFullName,
        baselearner_name: str,
    ):
        for bl in self._trained_stage_baselearners[metalearner_stage]:
            if bl.stage_bl.name == baselearner_name:
                return bl
        raise Exception("There isn't baselearner {}".format(baselearner_name))

    def _create_metalearner_wrap(
        self,
        metalearner_name: str,
        bls: Dict[MLStageFullName, str],
    ) -> MetaLearnerWrapper:
        """Create of best metalearner wrapper.

        Args:
            metalearner_name: Name.
            bls: Mapping metalearner stage to baselearner name.

        Returns:
            Best metalearner wrap.

        """
        ml_wrap_name = "__ML__{ML}".format(ML=metalearner_name)

        ml_wrap: Optional[MetaLearnerWrapper] = None
        if metalearner_name == "TLearner":
            ocl = self._get_trained_bl(("outcome_control"), bls[("outcome_control",)]).stage_bl
            otl = self._get_trained_bl(("outcome_treatment",), bls[("outcome_treatment",)]).stage_bl

            ml_wrap = MetaLearnerWrapper(
                name=ml_wrap_name,
                klass=TLearner,
                params={"control_learner": ocl, "treatment_learner": otl},
            )
        elif metalearner_name == "XLearner":
            ocl = self._get_trained_bl(("outcome_control",), bls[("outcome_control",)]).stage_bl
            otl = self._get_trained_bl(("outcome_treatment",), bls[("outcome_treatment",)]).stage_bl
            pl = self._get_trained_bl(("propensity",), bls[("propensity",)]).stage_bl
            ecl = self._get_trained_bl(
                ("outcome_treatment", "effect_control"), bls[("outcome_treatment", "effect_control")]
            ).stage_bl
            etl = self._get_trained_bl(
                ("outcome_control", "effect_treatment"),
                bls[("outcome_control", "effect_treatment")],
            ).stage_bl

            ml_wrap = MetaLearnerWrapper(
                name=ml_wrap_name,
                klass=XLearner,
                params={
                    "outcome_learners": [ocl, otl],
                    "effect_learners": [ecl, etl],
                    "propensity_learner": pl,
                },
            )
        else:
            raise Exception()

        return ml_wrap

    def __default_learners(
        self,
        lin_params: Optional[Dict[str, Any]] = None,
        tab_params: Optional[Dict[str, Any]] = None,
    ) -> List[BaseLearnerWrapper]:
        """Predefined baselearners."""
        default_lin_params = {"cpu_limit": self.cpu_limit}
        default_tab_params = {
            "timeout": None,
            "cpu_limit": self.cpu_limit,
            "gpu_ids": self.gpu_ids,
        }
        return [
            BaseLearnerWrapper(
                name="__Linear__",
                klass=uplift_utils.create_linear_automl,
                params=default_lin_params if lin_params is None else lin_params,
            ),
            BaseLearnerWrapper(
                name="__Tabular__",
                klass=TabularAutoML,
                params=default_tab_params if tab_params is None else tab_params,
            ),
        ]



"""AutoMLWhitebox for tabular datasets."""

import warnings

from copy import copy
from copy import deepcopy
from typing import Optional
from typing import Tuple
from typing import Union

import autowoe
import numpy as np

from pandas import DataFrame

from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..validation.base import TrainValidIterator
from .base import TabularMLAlgo


WbModel = Union[autowoe.AutoWoE, autowoe.ReportDeco]


class WbMLAlgo(TabularMLAlgo):
    """WhiteBox - scorecard model.

    https://github.com/AILab-MLTools/AutoMLWhitebox

    default_params:

       - monotonic: bool
           Global condition for monotonic constraints.
           If ``True``, then only monotonic binnings will be built.
           You can pass values to the ``.fit``
           method that change this condition
           separately for each feature.
       - max_bin_count: int
           Global limit for the number of bins. Can be specified for every
           feature in .fit
       - select_type: ``None`` or ``int``
           The type to specify the primary feature selection.
           If the type is an integer, then we select the number
           of features indicated by this number
           (with the best `feature_importance`).
           If the value is ``None``, we leave only features
           with ``feature_importance`` greater than ``0``.
       - pearson_th:  0 < pearson_th < 1
           Threshold for feature selection by correlation. All features with
           the absolute value of correlation coefficient greater then
           pearson_th will be discarded.
       - metric_th: .5 < metric_th < 1
           Threshold for feature selection by one-dimensional AUC.
           WoE with AUC < metric_th will be discarded.
       - vif_th: vif_th > 0
           Threshold for feature selection by VIF. Features with VIF > vif_th
           are iteratively discarded one by one, then VIF is recalculated
           until all VIFs are less than vif_th.
       - imp_th: real >= 0
           Threshold for feature selection by feature importance
       - th_const:
           Threshold, which determines that the feature is constant.
           If the number of valid values is greater than the threshold, then
           the column is not constant. For float, the number of
           valid values will be calculated as the sample size * th_const
       - force_single_split: bool
           In the tree parameters, you can set the minimum number of
           observations in the leaf. Thus, for some features,
           splitting for 2 beans at least will be impossible.
           If you specify that ``force_single_split = True``,
           it means that 1 split will be created for the feature,
           if the minimum bin size is greater than th_const.
       - th_nan: int >= 0
           Threshold, which determines that WoE values are calculated to NaN.
       - th_cat: int >= 0
           Threshold, which determines which categories are small.
       - woe_diff_th: float = 0.01
           The option to merge NaNs and rare categories with another bin,
           if the difference in WoE is less than woe_diff_th.
       - min_bin_size: int > 1, 0 < float < 1
           Minimum bin size when splitting.
       - min_bin_mults: list of floats > 1
           If minimum bin size is specified, you can specify a list to check
           if large values work better, for example: [2, 4].
       - min_gains_to_split: list of floats >= 0
           min_gain_to_split values that will be
           iterated to find the best split.
       - metric_tol: 1e-5 <= metric_tol <=1e-2
           Metric tolerance. You can lower the metric_tol value from the maximum
           to make the model simpler.
       - cat_alpha: float > 0
           Regularizer for category encoding.
       - cat_merge_to: str
           The way of WoE values filling in the test sample for categories
           that are not in the training sample.
           Values - 'to_nan', 'to_woe_0', 'to_maxfreq', 'to_maxp', 'to_minp'
       - nan_merge_to: str
           The way of WoE values filling on the test sample for real NaNs,
           if they are not included in their group.
           Values - 'to_woe_0', 'to_maxfreq', 'to_maxp', 'to_minp'
       - oof_woe: bool
           Use OOF or standard encoding for WOE.
       - n_folds: int
           Number of folds for feature selection / encoding, etc.
       - n_jobs: int > 0
           Number of CPU cores to run in parallel.
       - l1_base_step: real > 0
           Grid size in l1 regularization
       - l1_exp_step: real > 1
           Grid scale in l1 regularization
       - population_size: None, int > 0
           Feature selection type in the selector.
           If the value is ``None`` then L1 boost is used.
           If ``int`` is specified, then a standard step will be used for
           the number of random subsamples indicated by this value.
           Can be generalized to genetic algorithm.
       - feature_groups_count: int > 0
           The number of groups in the genetic algorithm.
           Its effect is visible only when population_size > 0
       - imp_type: str
           Feature importances type. Feature_imp and perm_imp are available.
           It is used to sort the features at the first and at the final
           stage of feature selection.
       - regularized_refit: bool
           Use regularization at the time of model refit. Otherwise, we have
           a statistical model.
       - p_val: 0 < p_val <= 1
           When training a statistical model, do backward selection
           until all p-values of the model's coefficient are
       - verbose: int 0-3
           Verbosity level

    freeze_defaults:
        - ``True`` :  params can be rewritten depending on dataset.
        - ``False``:  params can be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "WhiteBox"

    _default_params = {
        "monotonic": False,
        "max_bin_count": 5,
        "select_type": None,
        "pearson_th": 0.9,
        "metric_th": 0.505,
        "vif_th": 10.0,
        "imp_th": 0,
        "th_const": 32,
        "force_single_split": True,
        "th_nan": 0.01,
        "th_cat": 0.005,
        "woe_diff_th": 0.01,
        "min_bin_size": 0.01,
        "cat_alpha": 100,
        "cat_merge_to": "to_woe_0",
        "nan_merge_to": "to_woe_0",
        "oof_woe": True,
        "n_folds": 6,
        "n_jobs": 4,
        "l1_grid_size": 20,
        "l1_exp_scale": 6,
        "imp_type": "feature_imp",
        "regularized_refit": False,
        "p_val": 0.05,
        "report": False,
        "verbose": 0,
    }

    _report_on_inference = False

    def _infer_params(self) -> Tuple[dict, bool, dict]:

        params = deepcopy(self.params)
        report = params.pop("report")
        fit_params = params.pop("fit_params")
        self._report_on_inference = report
        return params, report, fit_params

    def fit_predict(self, train_valid_iterator: TrainValidIterator, **kwargs) -> NumpyDataset:  # noqa: D102

        self._dataset_fit_params = kwargs

        return super().fit_predict(train_valid_iterator)

    def _include_target(self, dataset: PandasDataset, include_group: bool = False) -> Tuple[DataFrame, Optional[str]]:

        df = dataset.data.copy()
        if dataset.target is not None:
            df["__TARGET__"], _ = self.task.losses["lgb"].fw_func(dataset.target.values, None)
        group_kf = None

        if include_group and dataset.group is not None:
            assert "__GROUP__" not in dataset.features, "__GROUP__ is not valid column name for WhiteBox"
            df["__GROUP__"] = dataset.group.values
            group_kf = "__GROUP__"

        return df, group_kf

    def fit_predict_single_fold(self, train: PandasDataset, valid: PandasDataset) -> Tuple[WbModel, np.ndarray]:
        """Implements training and prediction on single fold.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Tuple (model, predicted_values).

        """
        params, report, fit_params = self._infer_params()

        assert train.task.name == "binary", "Only binary task is supported"
        assert "__TARGET__" not in train.features, "__TARGET__ is not valid column name for WhiteBox"
        if train.weights is not None:
            warnings.warn("Weights are ignored at the moment", UserWarning, stacklevel=2)

        train_df, group_kf = self._include_target(train, True)

        roles = train.roles
        mapping = {"Category": "cat", "Numeric": "real"}
        features_type = {x: mapping[roles[x].name] for x in roles}

        valid_df = None
        if train is not valid:
            valid_df, _ = self._include_target(valid, False)

        model = autowoe.AutoWoE(**params)

        if report:
            model = autowoe.ReportDeco(model)

        kwargs = copy(self._dataset_fit_params)
        kwargs["validation"] = valid_df
        kwargs = {**kwargs, **fit_params}

        model.fit(train_df, target_name="__TARGET__", group_kf=group_kf, features_type=features_type, **kwargs)

        if train is valid:
            valid_df = train_df

        val_pred = model.predict_proba(valid_df)
        val_pred = self.task.losses["lgb"].bw_func(val_pred)

        return model, val_pred

    def predict_single_fold(self, model: WbModel, dataset: PandasDataset) -> np.ndarray:
        """Predict target values for dataset.

        Args:
            model: WhiteBox model
            dataset: Test dataset.

        Returns:
            Predicted target values.

        """
        args = []
        if self.params["report"]:
            args = [self._report_on_inference]

        df, _ = self._include_target(dataset, False)
        pred = self.task.losses["lgb"].bw_func(model.predict_proba(df, *args))

        return pred

    def fit(self, train_valid: TrainValidIterator):
        """Just to be compatible with ImportanceEstimator.

        Args:
            train_valid: classic cv iterator.

        """
        self.fit_predict(train_valid)

    def predict(self, dataset: PandasDataset, report: bool = False) -> NumpyDataset:
        """Predict on new dataset.

        Args:
            dataset: Dataset.
            report: Flag to generate report.

        Returns:
            Dataset with predictions.

        """
        self._report_on_inference = report
        return super().predict(dataset)



"""Tools to configure time utilization."""

import logging

from copy import deepcopy
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Sequence
from typing import Type
from typing import Union

from ...automl.base import AutoML
from ...automl.blend import BestModelSelector
from ...automl.blend import Blender
from ...automl.presets.base import AutoMLPreset
from ...dataset.base import LAMLDataset
from ...dataset.utils import concatenate
from ...ml_algo.base import MLAlgo
from ...pipelines.ml.base import MLPipeline
from ...tasks import Task
from ...utils.logging import set_stdout_level
from ...utils.logging import verbosity_to_loglevel
from ...utils.timer import PipelineTimer

from joblib import dump


logger = logging.getLogger(__name__)


class MLAlgoForAutoMLWrapper(MLAlgo):
    """Wrapper to apply blender to list of automl's."""

    @classmethod
    def from_automls(cls, automl: Union[AutoML, Sequence[AutoML]]):
        """Constructs automls.

        Args:
            automl: One AutoML or list of AutoML objects.

        Returns:
            MLAlgo.

        """
        ml_algo = cls()
        ml_algo.models.append(automl)

        return ml_algo

    def fit_predict(self, *args, **kwargs) -> LAMLDataset:
        raise NotImplementedError

    def predict(self, *args, **kwargs) -> LAMLDataset:
        raise NotImplementedError


class MLPipeForAutoMLWrapper(MLPipeline):
    """Wrapper to apply blender to list of automls."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ml_algos = self._ml_algos

    @classmethod
    def from_automl(cls, automl: AutoML):
        ml_pipe = cls([MLAlgoForAutoMLWrapper.from_automls(automl)])

        return ml_pipe

    @classmethod
    def from_blended(cls, automls: Sequence[AutoML], blender: Blender):
        ml_pipe = cls(
            [
                MLAlgoForAutoMLWrapper.from_automls(automls),
            ]
        )
        ml_pipe.blender = blender

        return ml_pipe


class TimeUtilization:
    """Class that helps to utilize given time to :class:`~lightautoml.automl.presets.base.AutoMLPreset`.

    Useful to calc benchmarks and compete
    It takes a list of config files as input and runs it until a time limit is exceeded.
    If time left - it can perform multistart on same configs with new random state.
    In best case - blend different configurations of single preset.
    In worst case - averaging multiple automl's with different states.

    Note:
        Basic usage.

        >>> ensembled_automl = TimeUtilization(TabularAutoML, Task('binary'),
        >>>     timeout=3600, configs_list=['cfg0.yml', 'cfg1.yml'])

        Then ``.fit_predict`` and predict can be
        called like usual :class:`~lightautoml.automl.base.AutoML` class.


    Args:
        automl_factory: One of presets.
        task: Task to solve.
        timeout: Timeout in seconds.
        memory_limit: Memory limit that are passed to each automl.
        cpu_limit: Cpu limit that that are passed to each automl.
        gpu_ids: Gpu_ids that are passed to each automl.
        verbose: Controls the verbosity: the higher, the more messages.
            <1  : messages are not displayed;
            >=1 : the computation process for layers is displayed;
            >=2 : the information about folds processing is also displayed;
            >=3 : the hyperparameters optimization process is also displayed;
            >=4 : the training process for every algorithm is displayed;
        timing_params: Timing_params level that are passed to each automl.
        configs_list: List of str path to configs files.
        inner_blend: Blender instance to blend automl's with same configs
            and different random state.
        outer_blend: Blender instance to blend averaged by random_state
            automl's with different configs.
        drop_last: Usually last automl will be stopped with timeout.
            Flag that defines if we should drop it from ensemble
        return_all_predictions: Skip blend and return all model predictions
        max_runs_per_config: Maximum number of multistart loops.
        random_state_keys: Params of config that used as
            random state with initial values. If ``None`` - search for
            `random_state` key in default config of preset.
            If not found - assume, that seeds are not fixed
            and each run is random by default. For example
            ``{'reader_params': {'random_state': 42}, 'gbm_params': {'default_params': {'seed': 42}}}``
        random_state: initial random seed, that will be
            set in case of search in config.
        **kwargs: Additional params.

    """

    def __init__(
        self,
        automl_factory: Type[AutoMLPreset],
        task: Task,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = None,
        timing_params: Optional[dict] = None,
        configs_list: Optional[Sequence[str]] = None,
        inner_blend: Optional[Blender] = None,
        outer_blend: Optional[Blender] = None,
        drop_last: bool = True,
        return_all_predictions: bool = False,
        max_runs_per_config: int = 5,
        random_state_keys: Optional[dict] = None,
        random_state: int = 42,
        **kwargs,
    ):
        self.automl_factory = automl_factory
        self.task = task
        self.timeout = timeout
        self.memory_limit = memory_limit
        self.cpu_limit = cpu_limit
        self.gpu_ids = gpu_ids

        self.timing_params = timing_params
        if timing_params is None:
            self.timing_params = {}

        self.configs_list = configs_list
        if configs_list is None:
            self.configs_list = [None]

        self.max_runs_per_config = max_runs_per_config

        self.random_state_keys = random_state_keys
        if random_state_keys is None:
            self.random_state_keys = self._search_for_states(automl_factory, random_state)

        self.inner_blend = inner_blend
        if inner_blend is None:
            self.inner_blend = BestModelSelector()

        self.outer_blend = outer_blend
        if outer_blend is None:
            self.outer_blend = BestModelSelector()
        self.drop_last = drop_last
        self.return_all_predictions = return_all_predictions
        self.kwargs = kwargs

    def _search_for_key(self, config, key, value: int = 42) -> dict:

        d = {}

        if key in config:
            d[key] = value

        for k in config:
            if type(config[k]) is dict:
                s = self._search_for_key(config[k], key, value)
                if len(s) > 0:
                    d[k] = s
        return d

    def _search_for_states(self, automl_factory: Type[AutoMLPreset], random_state: int = 42) -> dict:

        config = automl_factory.get_config()
        random_states = self._search_for_key(config, "random_state", random_state)

        return random_states

    def _get_upd_states(self, random_state_keys: dict, upd_value: int = 0) -> dict:

        d = {}

        for k in random_state_keys:
            if type(random_state_keys[k]) is dict:
                d[k] = self._get_upd_states(random_state_keys[k], upd_value)
            else:
                d[k] = random_state_keys[k] + upd_value

        return d

    def fit_predict(
        self,
        train_data: Any,
        roles: dict,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[Any] = None,
        valid_features: Optional[Sequence[str]] = None,
        verbose: int = 0,
        log_file: str = None,
        path_to_save: Optional[str] = None,
    ) -> LAMLDataset:
        """Fit and get prediction on validation dataset.

        Almost same as :meth:`lightautoml.automl.base.AutoML.fit_predict`.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names, if can't
                be inferred from `train_data`.
            cv_iter: Custom cv-iterator. For example,
                :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset features
                if cannot be inferred from `valid_data`.
            verbose: Verbose.
            log_file: Log filename.
            path_to_save: The path that joblib will use to save the model after fit stage is completed. Use *.joblib format.

        Returns:
            Dataset with predictions. Call ``.data`` to get predictions array.

        """
        set_stdout_level(verbosity_to_loglevel(verbose))

        logger.info("Start automl \x1b[1mutilizator\x1b[0m with listed constraints:")
        logger.info(f"- time: {self.timeout:.2f} seconds")
        logger.info(f"- CPU: {self.cpu_limit} cores")
        logger.info(f"- memory: {self.memory_limit} GB\n")
        logger.info("\x1b[1mIf one preset completes earlier, next preset configuration will be started\x1b[0m\n")

        timer = PipelineTimer(self.timeout, **self.timing_params).start()
        history = []

        amls = [[] for _ in range(len(self.configs_list))]
        aml_preds = [[] for _ in range(len(self.configs_list))]
        class_mapping = None
        n_ms = 0
        n_cfg = 0
        upd_state_val = 0
        flg_continute = True
        # train automls one by one while timer is ok
        while flg_continute:
            n_ms += 1

            logger.info("=" * 50)

            for n_cfg, config in enumerate(self.configs_list):
                random_states = self._get_upd_states(self.random_state_keys, upd_state_val)
                random_states["general_params"] = {"return_all_predictions": False}
                upd_state_val += 1

                logger.info(f"Start {n_cfg} automl preset configuration:")
                logger.info("\x1b[1m{}\x1b[0m, random state: {}".format(config.split("/")[-1], random_states))

                cur_kwargs = self.kwargs.copy()
                for k in random_states.keys():
                    if k in self.kwargs:
                        logger.info3("Found {} in kwargs, need to combine".format(k))
                        random_states[k] = {**cur_kwargs[k], **random_states[k]}
                        del cur_kwargs[k]
                        logger.info3("Merged variant for {} = {}".format(k, random_states[k]))

                automl = self.automl_factory(
                    self.task,
                    timer.time_left,
                    memory_limit=self.memory_limit,
                    cpu_limit=self.cpu_limit,
                    gpu_ids=self.gpu_ids,
                    timing_params=self.timing_params,
                    config_path=config,
                    **random_states,
                    **cur_kwargs,
                )

                val_pred = automl.fit_predict(
                    train_data,
                    roles,
                    train_features,
                    cv_iter,
                    valid_data,
                    valid_features,
                    verbose=verbose,
                    log_file=log_file,
                )

                current_class_mapping = automl.reader.class_mapping

                if class_mapping is None:
                    class_mapping = current_class_mapping
                else:
                    assert (
                        class_mapping == current_class_mapping
                    ), "class_mappings are different for different AutoML for some reason."

                logger.info("=" * 50)

                amls[n_cfg].append(MLPipeForAutoMLWrapper.from_automl(automl))
                aml_preds[n_cfg].append(val_pred)

                history.append(timer.time_spent - sum(history))
                if timer.time_left < (sum(history) / len(history)) or upd_state_val >= (
                    self.max_runs_per_config * len(self.configs_list)
                ):
                    flg_continute = False
                    break

        # usually last model will be not complete due to timeout.
        # Maybe it's better to remove it from inner blend, which is typically just mean of models
        if n_ms > 1 and self.drop_last:
            amls[n_cfg].pop()
            aml_preds[n_cfg].pop()

        # prune empty algos
        amls = [x for x in amls if len(x) > 0]
        aml_preds = [x for x in aml_preds if len(x) > 0]

        # blend - first is inner blend - we blend same config with different states
        inner_pipes = []
        inner_preds = []

        for preds, pipes in zip(aml_preds, amls):
            inner_blend = deepcopy(self.inner_blend)
            val_pred, inner_pipe = inner_blend.fit_predict(preds, pipes, class_mapping=class_mapping)
            inner_pipe = [x.ml_algos[0].models[0] for x in inner_pipe]

            inner_preds.append(val_pred)
            inner_pipes.append(MLPipeForAutoMLWrapper.from_blended(inner_pipe, inner_blend))

        # outer blend - blend of blends
        if not self.return_all_predictions:
            val_pred, self.outer_pipes = self.outer_blend.fit_predict(
                inner_preds, inner_pipes, class_mapping=class_mapping
            )
        else:
            val_pred = concatenate(inner_preds)
            self.outer_pipes = inner_pipes

        self.targets_order = sorted(class_mapping, key=class_mapping.get, reverse=False) if class_mapping else None

        # saving automl model with joblib
        if path_to_save is not None:
            # There is 1 parameter for model save:
            # "path_to_save" - name of model for saving.

            dump_name = path_to_save if path_to_save.endswith(".joblib") else f"{path_to_save}.joblib"
            dump(self, dump_name, compress=0)

        return val_pred

    def predict(
        self,
        data: Any,
        features_names: Optional[Sequence[str]] = None,
        return_all_predictions: Optional[bool] = None,
        **kwargs,
    ) -> LAMLDataset:
        """Get dataset with predictions.

        Almost same as :meth:`lightautoml.automl.base.AutoML.predict`
        on new dataset, with additional features.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`. For example,
              ``{'data': X...}``. In this case roles are optional,
              but `train_features` and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
                if cannot be inferred from `train_data`.
            return_all_predictions: bool - skip blending phase
            **kwargs: Other params.

        Returns:
            Dataset with predictions.

        """
        if return_all_predictions is None or self.return_all_predictions:
            return_all_predictions = self.return_all_predictions

        outer_preds = []

        for amls_pipe in self.outer_pipes:

            inner_preds = []
            # TODO: Maybe refactor?
            for automl in amls_pipe.ml_algos[0].models[0]:
                inner_pred = automl.predict(data, features_names, **kwargs)
                inner_preds.append(inner_pred)

            outer_pred = amls_pipe.blender.predict(inner_preds)
            outer_preds.append(outer_pred)

        # pred = self.outer_blend.predict(outer_preds)

        if not return_all_predictions:
            pred = self.outer_blend.predict(outer_preds)
        else:
            pred = concatenate(outer_preds)

        return pred



"""Data processing utils."""
from collections import Counter
from collections import defaultdict
from random import shuffle
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
import torch

from torch import nn
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.data import Sampler


class LengthDataset(Dataset):
    """Default dict like PyTorch dataset, with additional info about lenght of sequence.

    Args:
        tokens: List of tokens number, numbering from zero to dictionary size.
        targets: Array with targets.

    """

    def __init__(self, tokens: List[List[int]], targets: np.ndarray):
        self.tokens = tokens
        self.targets = targets

    def __getitem__(self, idx):
        return {
            "text": self.tokens[idx].clone().detach(),
            "len": len(self.tokens[idx]),
            "target": torch.from_numpy(self.targets[idx]),
        }

    def __len__(self):
        return len(self.targets)


class BySequenceLengthSampler(Sampler):
    """PyTorch sampler with binning by sequnce length.

    Args:
        data_source: Dataset in dict like format, with key `'len'`.
            The values corresponding to this key should contain
            the length of the sentence (i.e. number of tokens).
        bucket_boundaries: Binning boundaries. The
        batch_size: Number of elements from bin used for batching.
        drop_last: Set to `True` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size,
            else leave the remaining part.
        shuffle: Shuffle flag. If `True` the internally
            and externally bins will be shuffled.

    """

    def __init__(
        self,
        data_source: LengthDataset,
        bucket_boundaries: List[int],
        batch_size: int = 64,
        drop_last: bool = True,
        shuffle: bool = True,
    ):
        self.data_source = data_source

        ind_n_len = []

        for i, d in enumerate(self.data_source):
            ind_n_len.append((i, d["len"]))

        self.bucket_boundaries = bucket_boundaries
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.ind_n_len = ind_n_len
        self.boundaries = list(self.bucket_boundaries)
        self.buckets_min = torch.tensor([np.iinfo(np.int32).min] + self.boundaries)
        self.buckets_max = torch.tensor(self.boundaries + [np.iinfo(np.int32).max])
        self.boundaries = torch.tensor(self.boundaries)
        self.shuffle = shuffle

    def __iter__(self):
        data_buckets = defaultdict(list)
        for i, slen in self.ind_n_len:
            pid = self.to_bucket(slen)
            data_buckets[pid].append(i)

        for k in data_buckets.keys():
            data_buckets[k] = torch.tensor(data_buckets[k])

        iter_list = []
        for k in data_buckets.keys():
            t = self.shuffle_tensor(data_buckets[k])
            batch = torch.split(t, self.batch_size, dim=0)

            if self.drop_last and len(batch[-1]) != self.batch_size:
                batch = batch[:-1]

            iter_list += batch

        if self.shuffle:
            shuffle(iter_list)

        for i in iter_list:
            yield i.numpy().tolist()

    def __len__(self):
        return len(self.data_source) // self.batch_size + 1

    def to_bucket(self, seq_length):  # noqa D102
        valid_buckets = (seq_length >= self.buckets_min) * (seq_length < self.buckets_max)
        bucket_id = torch.nonzero(valid_buckets, as_tuple=True)[0].item()

        return bucket_id

    def shuffle_tensor(self, t):
        """Shuffle tensor along lenght."""
        if self.shuffle:
            return t[torch.randperm(len(t))]
        else:
            return t


def pad_max_len(batch: List[Dict[str, Any]]) -> Dict[str, Any]:
    """Collate function, that pads sequences to maximum length of text in batch.

    Args:
        batch: List of dictionaries from LengthDataset.

    Returns:
        Dict of vectors of padded texts, targets.

    """
    texts = []
    lens = []
    targets = []
    for d in batch:
        texts.append(d["text"])
        lens.append(d["len"])
        targets.append(d["target"])

    return {
        "text": pad_sequence(texts, batch_first=True, padding_value=0),
        "len": lens,
        "target": torch.stack(targets),
    }


def get_tokenized(data: pd.DataFrame, tokenized_col: str, tokenizer: Callable) -> List[List[str]]:
    """Get tokenized column.

    Args:
        data: Dataset with text column.
        tokenized_col: Column used to tokenized.
        tokenizer: Tokenizer function str->List[str].

    Returns:
        Tokenized column.

    """
    tokenized = []
    for sent in data[tokenized_col]:
        tokenized.append(tokenizer(sent.lower()))

    return tokenized


def get_vocab(tokenized: List[List[str]], max_len_vocab: int) -> Tuple[Dict[str, int], Dict[int, str]]:
    """Get vocabulary dataset.

    Args:
        tokenized: Tokenized strings.
        max_len_vocab: Vocabulary size.

    Returns:
        Word to index, index to word.

    """
    counter = Counter()
    for sent in tokenized:
        counter.update(sent)
    max_len_vocab = len(counter) if max_len_vocab == -1 else max_len_vocab
    max_len_vocab = min(max_len_vocab, len(counter))
    word_to_id = {"<PAD>": 0, "<START>": 1, "<UNK>": 2}
    word_to_id.update({v: i + 3 for i, (v, _) in enumerate(counter.most_common(max_len_vocab))})
    id_to_word = {w: k for k, w in word_to_id.items()}

    return word_to_id, id_to_word


def get_embedding_matrix(id_to_word: Dict[int, str], embedder: Any, embedder_dim: int) -> torch.FloatTensor:
    """Create embedding matrix from lookup table and (optionaly) embedder.

    Args:
        id_to_word: Dictionary token-id to word.
        embedder: Lookup table with embeddings.
        embedder_dim: Dimetion of embeddings.

    Returns:
        Weight matrix of embedding layer.

    """
    weights_matrix = np.random.normal(scale=0.6, size=(len(id_to_word), embedder_dim))
    if embedder is not None:
        for i in range(len(id_to_word)):
            word = id_to_word[i]
            try:
                emb = embedder[word]
            except KeyError or TypeError:
                continue
            if sum(emb) != 0:
                weights_matrix[i] = emb

    weights_matrix = torch.from_numpy(weights_matrix)

    return weights_matrix


def map_tokenized_to_id(
    tokenized: List[List[str]],
    word_to_id: "lightautoml.addons.interpretation.utils.WrappedVocabulary",  # noqa F821
    min_k: int,
) -> List[torch.LongTensor]:
    """Mapping from word dataset to tokenized dataset.

    Args:
        tokenized: Dataset of words.
        word_to_id: Dictionary word to token-id.
        min_k: Parameter for minumum sequence length (used for padding).

    Returns:
        Dataset with token ids.

    """
    dataset = []
    for sent in tokenized:
        sent_list = [word_to_id["<START>"]]
        # word_to_id is also callable
        sent_list.extend(map(word_to_id, sent))
        pad_tokens = max(1, min_k - len(sent_list))
        sent_list.extend([word_to_id["<PAD>"]] * pad_tokens)
        dataset.append(torch.Tensor(sent_list).long())

    return dataset


def get_len_dataset(tokenized: List[List[str]], target: np.ndarray) -> LengthDataset:
    """Get length dataset.

    Args:
        tokenized: Tokenized strings.
        target: Target values.

    Returns:
        Length Dataset.

    Raises:
        ValueError: Length missmatch of target and tokenized.

    """
    if len(tokenized) != len(target):
        raise ValueError("Missmatch of lengths tokenized ({}) and target ({})".format(len(tokenized), len(target)))
    return LengthDataset(tokenized, target)


def get_len_dataloader(
    dataset: LengthDataset,
    batch_size: int = 1,
    boundaries: Optional[List[int]] = None,
    mode: str = "train",
) -> torch.utils.data.DataLoader:
    """Get len dataloader.

    Args:
        dataset: Lenght dataset.
        batch_size: Size of batch.
        boundaries: List of binning boundaries.
        mode: Stage of dataloader.

    Returns:
        Dataloader for L2X.

    """
    if mode == "train":
        sampler = BySequenceLengthSampler(dataset, boundaries, batch_size, drop_last=True)
        dataloader = DataLoader(dataset, batch_size=1, batch_sampler=sampler, collate_fn=pad_max_len)
    elif mode == "valid":
        sampler = BySequenceLengthSampler(dataset, boundaries, batch_size, drop_last=False, shuffle=False)
        dataloader = DataLoader(dataset, batch_size=1, batch_sampler=sampler, collate_fn=pad_max_len)
    elif mode == "test":
        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=pad_max_len)

    return dataloader


def create_emb_layer(weights_matrix=None, voc_size=None, embed_dim=None, trainable_embeds=True) -> torch.nn.Embedding:
    """Create initialized embedding layer.

    Args:
        weights_matrix: Weights of embedding layer.
        voc_size: Size of vocabulary.
        embed_dim: Size of embeddings.
        trainable_embeds: To optimize layer when training model.

    Returns:
        Initialized embedding layer.

    """
    assert (weights_matrix is not None) or (
        voc_size is not None and embed_dim is not None
    ), "Please define anything: weights_matrix or voc_size & embed_dim"

    if weights_matrix is not None:
        voc_size, embed_dim = weights_matrix.size()
    emb_layer = nn.Embedding(voc_size, embed_dim)
    if weights_matrix is not None:
        emb_layer.load_state_dict({"weight": weights_matrix})
    if not trainable_embeds:
        emb_layer.weight.requires_grad = False

    return emb_layer



"""Linear models for tabular datasets."""

import logging

from copy import copy
from copy import deepcopy
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np

from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lasso
from sklearn.linear_model import LogisticRegression

from ..dataset.np_pd_dataset import PandasDataset
from ..validation.base import TrainValidIterator
from .base import TabularDataset
from .base import TabularMLAlgo
from .torch_based.linear_model import TorchBasedLinearEstimator
from .torch_based.linear_model import TorchBasedLinearRegression
from .torch_based.linear_model import TorchBasedLogisticRegression


logger = logging.getLogger(__name__)

LinearEstimator = Union[LogisticRegression, ElasticNet, Lasso]


class LinearLBFGS(TabularMLAlgo):
    """LBFGS L2 regression based on torch.

    default_params:

        - cs: List of regularization coefficients.
        - max_iter: Maximum iterations of L-BFGS.
        - tol: The tolerance for the stopping criteria.
        - early_stopping: Maximum rounds without improving.

    freeze_defaults:

        - ``True`` :  params can be rewritten depending on dataset.
        - ``False``:  params can be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "LinearL2"

    _default_params = {
        "tol": 1e-6,
        "max_iter": 100,
        "cs": [
            1e-5,
            5e-5,
            1e-4,
            5e-4,
            1e-3,
            5e-3,
            1e-2,
            5e-2,
            1e-1,
            5e-1,
            1,
            5,
            10,
            50,
            100,
            500,
            1000,
            5000,
            10000,
            50000,
            100000,
        ],
        "early_stopping": 2,
    }

    def _infer_params(self) -> TorchBasedLinearEstimator:

        params = copy(self.params)
        params["loss"] = self.task.losses["torch"].loss
        params["metric"] = self.task.losses["torch"].metric_func

        if self.task.name in ["multilabel"]:
            model = TorchBasedLogisticRegression(output_size=self.n_classes, multilabel=True, **params)
        elif self.task.name in ["binary", "multiclass"]:
            model = TorchBasedLogisticRegression(output_size=self.n_classes, **params)
        elif self.task.name == "reg":
            model = TorchBasedLinearRegression(output_size=1, **params)
        elif self.task.name == "multi:reg":
            model = TorchBasedLinearRegression(output_size=self.n_classes, **params)
        else:
            raise ValueError("Task not supported")

        return model

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:  # noqa: D102

        suggested_params = copy(self.default_params)
        train = train_valid_iterator.train
        suggested_params["categorical_idx"] = [
            n for (n, x) in enumerate(train.features) if train.roles[x].name == "Category"
        ]

        suggested_params["embed_sizes"] = ()
        if len(suggested_params["categorical_idx"]) > 0:
            suggested_params["embed_sizes"] = (
                train.data[:, suggested_params["categorical_idx"]].max(axis=0).astype(np.int32) + 1
            )

        suggested_params["data_size"] = train.shape[1]

        return suggested_params

    def fit_predict_single_fold(
        self, train: TabularDataset, valid: TabularDataset
    ) -> Tuple[TorchBasedLinearEstimator, np.ndarray]:
        """Train on train dataset and predict on holdout dataset.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Target predictions for valid dataset.

        """
        if type(train) is PandasDataset:
            train = train.to_numpy()
            valid = valid.to_numpy()

        model = self._infer_params()

        model.fit(
            train.data,
            train.target,
            train.weights,
            valid.data,
            valid.target,
            valid.weights,
        )

        val_pred = model.predict(valid.data)

        return model, val_pred

    def predict_single_fold(self, model: TorchBasedLinearEstimator, dataset: TabularDataset) -> np.ndarray:
        """Implements prediction on single fold.

        Args:
            model: Model uses to predict.
            dataset: ``NumpyDataset`` used for prediction.

        Returns:
            Predictions for input dataset.

        """
        pred = model.predict(dataset.data)

        return pred


class LinearL1CD(TabularMLAlgo):
    """Coordinate descent based on sklearn implementation."""

    _name: str = "LinearElasticNet"

    _default_params = {
        "tol": 1e-3,
        "max_iter": 100,
        "cs": [1e-4, 1e-3, 1e-2, 1e-1, 1, 10, 100, 1000, 10000, 100000, 1000000],
        "early_stopping": 2,
        "l1_ratios": (1,),
        "solver": "saga",
    }

    def _infer_params(
        self,
    ) -> Tuple[LinearEstimator, Sequence[float], Sequence[float], int]:

        params = copy(self.params)
        l1_ratios = params.pop("l1_ratios")
        early_stopping = params.pop("early_stopping")
        cs = params.pop("cs")

        if self.task.name in ["binary", "multiclass"]:

            if l1_ratios == (1,):
                model = LogisticRegression(warm_start=True, penalty="l1", **params)
            else:
                model = LogisticRegression(warm_start=True, penalty="elasticnet", **params)

        elif self.task.name == "reg":
            params.pop("solver")
            if l1_ratios == (1,):
                model = Lasso(warm_start=True, **params)
            else:
                model = ElasticNet(warm_start=True, **params)

        else:
            raise AttributeError("Task not supported")

        return model, cs, l1_ratios, early_stopping

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        """Get model parameters depending on dataset parameters.

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Parameters of model.

        """
        suggested_params = copy(self.default_params)
        task = train_valid_iterator.train.task

        assert "sklearn" in task.losses, "Sklearn loss should be defined"

        if task.name == "reg":
            # suggested_params['cs'] = list(map(lambda x: 1 / (2 * x), suggested_params['cs']))
            suggested_params["cs"] = [1 / (2 * i) for i in suggested_params["cs"]]

        return suggested_params

    def _predict_w_model_type(self, model, data):

        if self.task.name == "binary":
            pred = model.predict_proba(data)[:, 1]

        elif self.task.name == "reg":
            pred = model.predict(data)

        elif (self.task.name == "multiclass") or (self.task.name == "multilabel"):
            pred = model.predict_proba(data)

        else:
            raise ValueError("Task not supported")

        return pred

    def fit_predict_single_fold(
        self, train: TabularDataset, valid: TabularDataset
    ) -> Tuple[LinearEstimator, np.ndarray]:
        """Train on train dataset and predict on holdout dataset.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Target predictions for valid dataset.

        """
        if type(train) is PandasDataset:
            train = train.to_numpy()
            valid = valid.to_numpy()

        _model, cs, l1_ratios, early_stopping = self._infer_params()

        train_target, train_weight = self.task.losses["sklearn"].fw_func(train.target, train.weights)
        valid_target, valid_weight = self.task.losses["sklearn"].fw_func(valid.target, valid.weights)

        model = deepcopy(_model)

        best_score = -np.inf
        best_pred = None
        best_model = None

        metric = self.task.losses["sklearn"].metric_func

        for l1_ratio in sorted(l1_ratios, reverse=True):

            try:
                model.set_params(**{"l1_ratio": l1_ratio})
            except ValueError:
                pass

            model = deepcopy(_model)

            c_best_score = -np.inf
            c_best_pred = None
            c_best_model = None
            es = 0

            for n, c in enumerate(cs):

                try:
                    model.set_params(**{"C": c})
                except ValueError:
                    model.set_params(**{"alpha": c})

                model.fit(train.data, train_target, train_weight)

                if np.allclose(model.coef_, 0):
                    if n == (len(cs) - 1):
                        logger.info2(
                            "All model coefs are 0. Model with l1_ratio {0} is dummy".format(l1_ratio),
                            UserWarning,
                        )
                    else:
                        logger.debug("C = {0} all model coefs are 0".format(c))
                        continue

                pred = self._predict_w_model_type(model, valid.data)
                score = metric(valid_target, pred, valid_weight)

                logger.debug("C = {0}, l1_ratio = {1}, score = {2}".format(c, 1, score))

                # TODO: check about greater and equal
                if score >= c_best_score:
                    c_best_score = score
                    c_best_pred = deepcopy(pred)
                    es = 0
                    c_best_model = deepcopy(model)
                else:
                    es += 1

                if es >= early_stopping:
                    logger.debug("Early stopping..")
                    break

                if self.timer.time_limit_exceeded():
                    logger.info3("Time limit exceeded")
                    break

                # TODO: Think about is it ok to check time inside train loop?
                if (model.coef_ != 0).all():
                    logger.debug("All coefs are nonzero")
                    break

            if c_best_score >= best_score:
                best_score = c_best_score
                best_pred = deepcopy(c_best_pred)
                best_model = deepcopy(c_best_model)

            if self.timer.time_limit_exceeded():
                logger.info3("Time limit exceeded")
                break

        val_pred = self.task.losses["sklearn"].bw_func(best_pred)

        return best_model, val_pred

    def predict_single_fold(self, model: LinearEstimator, dataset: TabularDataset) -> np.ndarray:
        """Implements prediction on single fold.

        Args:
            model: Model uses to predict.
            dataset: Dataset used for prediction.

        Returns:
            Predictions for input dataset.

        """
        pred = self.task.losses["sklearn"].bw_func(self._predict_w_model_type(model, dataset.data))

        return pred



"""Reader and its derivatives."""

import logging

from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast
import warnings

import numpy as np
import pandas as pd

from pandas import DataFrame
from pandas import Series

from ..dataset.base import array_attr_roles
from ..dataset.base import valid_array_attributes
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import CategoryRole
from ..dataset.roles import ColumnRole
from ..dataset.roles import DatetimeRole
from ..dataset.roles import DropRole
from ..dataset.roles import NumericRole
from ..dataset.seq_np_pd_dataset import SeqNumpyPandasDataset
from ..dataset.utils import roles_parser
from ..tasks import Task
from .guess_roles import calc_category_rules
from .guess_roles import calc_encoding_rules
from .guess_roles import get_category_roles_stat
from .guess_roles import get_null_scores
from .guess_roles import get_numeric_roles_stat
from .guess_roles import rule_based_cat_handler_guess
from .guess_roles import rule_based_roles_guess
from .seq import IDSInd
from .seq import TopInd
from .utils import set_sklearn_folds


logger = logging.getLogger(__name__)

# roles, how it's passed to automl
RoleType = TypeVar("RoleType", bound=ColumnRole)
RolesDict = Dict[str, RoleType]

# how user can define roles
UserDefinedRole = Optional[Union[str, RoleType]]

UserDefinedRolesDict = Dict[UserDefinedRole, Sequence[str]]
UserDefinedRolesSequence = Sequence[UserDefinedRole]
UserRolesDefinition = Optional[Union[UserDefinedRole, UserDefinedRolesDict, UserDefinedRolesSequence]]

attrs_dict = dict(zip(array_attr_roles, valid_array_attributes))


class Reader:
    """Abstract Reader class.

    Abstract class for analyzing input data and creating inner
    :class:`~lightautoml.dataset.base.LAMLDataset` from raw data.
    Takes data in different formats as input,
    drop obviously useless features,
    estimates available size and returns dataset.

    Args:
        task: Task object
        *args: Not used.
        *kwargs: Not used.

    """

    def __init__(self, task: Task, *args: Any, **kwargs: Any):
        self.task = task
        self._roles = {}
        self._dropped_features = []
        self._used_array_attrs = {}
        self._used_features = []

    @property
    def roles(self) -> RolesDict:
        """Roles dict."""
        return self._roles

    @property
    def dropped_features(self) -> List[str]:
        """List of dropped features."""
        return self._dropped_features

    @property
    def used_features(self) -> List[str]:
        """List of used features."""
        return self._used_features

    @property
    def used_array_attrs(self) -> Dict[str, str]:
        """Dict of used array attributes."""
        return self._used_array_attrs

    def fit_read(
        self,
        train_data: Any,
        features_names: Optional[List[str]] = None,
        roles: UserRolesDefinition = None,
        **kwargs: Any,
    ):
        """Abstract function to get dataset with initial feature selection."""
        raise NotImplementedError

    def read(self, data: Any, features_names: Optional[List[str]], **kwargs: Any):
        """Abstract function to add validation columns."""
        raise NotImplementedError

    def upd_used_features(
        self,
        add: Optional[Sequence[str]] = None,
        remove: Optional[Sequence[str]] = None,
    ):
        """Updates the list of used features.

        Args:
            add: List of feature names to add or None.
            remove: List of feature names to remove or None.

        """
        curr_feats = set(self.used_features)
        if add is not None:
            curr_feats = curr_feats.union(add)
        if remove is not None:
            curr_feats = curr_feats - set(remove)
        self._used_features = list(curr_feats)

    @classmethod
    def from_reader(cls, reader: "Reader", **kwargs) -> "Reader":
        """Create reader for new data type from existed.

        Note - for now only Pandas reader exists, made for future plans.

        Args:
            reader: Source reader.
            **kwargs: Ignored as in the class itself.

        Returns:
            New reader.

        """
        new_reader = cls(reader.task, **kwargs)

        for attr in reader.__dict__:
            if attr[0] == "_":
                cls.__dict__[attr] = getattr(reader, attr)

        return new_reader

    def cols_by_type(self, col_type: str) -> List[str]:
        """Get roles names by it's type.

        Args:
            col_type: Column type, for example 'Text'.

        Returns:
            Array with column names.

        """
        names = []
        for col, role in self.roles.items():
            if role.name == col_type:
                names.append(col)

        return names


class PandasToPandasReader(Reader):
    """Pandas Reader.

    Reader to convert :class:`~pandas.DataFrame` to AutoML's :class:`~lightautoml.dataset.np_pd_dataset.PandasDataset`.
    Stages:

        - Drop obviously useless features.
        - Convert roles dict from user format to automl format.
        - Simple role guess for features without input role.
        - Create cv folds.
        - Create initial PandasDataset.
        - Optional: advanced guessing of role and handling types.


    Args:
        task: Task object.
        samples: Number of elements used when checking role type.
        max_nan_rate: Maximum nan-rate.
        max_constant_rate: Maximum constant rate.
        cv: CV Folds.
        random_state: Random seed.
        roles_params: dict of params of features roles. \
            Ex. {'numeric': {'dtype': np.float32}, 'datetime': {'date_format': '%Y-%m-%d'}}
            It's optional and commonly comes from config
        n_jobs: Int number of processes.
        advanced_roles: Param of roles guess (experimental, do not change).
        numeric_unique_rate: Param of roles guess (experimental, do not change).
        max_to_3rd_rate: Param of roles guess (experimental, do not change).
        binning_enc_rate: Param of roles guess (experimental, do not change).
        raw_decr_rate: Param of roles guess (experimental, do not change).
        max_score_rate: Param of roles guess (experimental, do not change).
        abs_score_val: Param of roles guess (experimental, do not change).
        drop_score_co: Param of roles guess (experimental, do not change).
        **kwargs: For now not used.

    """

    def __init__(
        self,
        task: Task,
        samples: Optional[int] = 100000,
        max_nan_rate: float = 0.999,
        max_constant_rate: float = 0.999,
        cv: int = 5,
        random_state: int = 42,
        roles_params: Optional[dict] = None,
        n_jobs: int = 4,
        # params for advanced roles guess
        advanced_roles: bool = True,
        numeric_unique_rate: float = 0.999,
        max_to_3rd_rate: float = 1.1,
        binning_enc_rate: float = 2,
        raw_decr_rate: float = 1.1,
        max_score_rate: float = 0.2,
        abs_score_val: float = 0.04,
        drop_score_co: float = 0.01,
        **kwargs: Any,
    ):
        super().__init__(task)
        self.samples = samples
        self.max_nan_rate = max_nan_rate
        self.max_constant_rate = max_constant_rate
        self.cv = cv
        self.random_state = random_state
        self.n_jobs = n_jobs

        self.roles_params = roles_params
        self.target = None
        if roles_params is None:
            self.roles_params = {}

        self.advanced_roles = advanced_roles
        self.advanced_roles_params = {
            "numeric_unique_rate": numeric_unique_rate,
            "max_to_3rd_rate": max_to_3rd_rate,
            "binning_enc_rate": binning_enc_rate,
            "raw_decr_rate": raw_decr_rate,
            "max_score_rate": max_score_rate,
            "abs_score_val": abs_score_val,
            "drop_score_co": drop_score_co,
        }

        self.params = kwargs
        self.class_mapping: Optional[Union[Mapping, Dict[str, Mapping]]] = None
        self._n_classes: Optional[int] = None

    def fit_read(
        self, train_data: DataFrame, features_names: Any = None, roles: UserDefinedRolesDict = None, **kwargs: Any
    ) -> PandasDataset:
        """Get dataset with initial feature selection.

        Args:
            train_data: Input data.
            features_names: Ignored. Just to keep signature.
            roles: Dict of features roles in format ``{RoleX: ['feat0', 'feat1', ...], RoleY: 'TARGET', ....}``.
            **kwargs: Can be used for target/group/weights.

        Returns:
            Dataset with selected features.

        """
        logger.info("\x1b[1mTrain data shape: {}\x1b[0m\n".format(train_data.shape))

        if roles is None:
            roles = {}
        # transform roles from user format {RoleX: ['feat0', 'feat1', ...], RoleY: 'TARGET', ....}
        # to automl format {'feat0': RoleX, 'feat1': RoleX, 'TARGET': RoleY, ...}
        parsed_roles = roles_parser(roles)
        # transform str role definition to automl ColumnRole
        attrs_dict = dict(zip(array_attr_roles, valid_array_attributes))

        for feat in parsed_roles:
            r = parsed_roles[feat]
            if isinstance(r, str):
                # get default role params if defined
                r = self._get_default_role_from_str(r)

            # check if column is defined like target/group/weight etc ...
            if r.name in attrs_dict:
                # defined in kwargs is rewritten.. TODO: Maybe raise warning if rewritten?
                # TODO: Think, what if multilabel or multitask? Multiple column target ..
                # TODO: Maybe for multilabel/multitask make target only available in kwargs??
                if ((self.task.name == "multi:reg") or (self.task.name == "multilabel")) and (
                    attrs_dict[r.name] == "target"
                ):
                    if attrs_dict[r.name] in kwargs:
                        kwargs[attrs_dict[r.name]].append(feat)
                        self._used_array_attrs[attrs_dict[r.name]].append(feat)
                    else:
                        kwargs[attrs_dict[r.name]] = [feat]
                        self._used_array_attrs[attrs_dict[r.name]] = [feat]
                else:
                    self._used_array_attrs[attrs_dict[r.name]] = feat
                    kwargs[attrs_dict[r.name]] = train_data.loc[:, feat]
                r = DropRole()

            # add new role
            parsed_roles[feat] = r

        assert "target" in kwargs, "Target should be defined"
        if self.task.name in ["multi:reg", "multilabel"]:
            kwargs["target"] = train_data.loc[:, kwargs["target"]]
        self.target = kwargs["target"].name if isinstance(kwargs["target"], pd.Series) else kwargs["target"].columns
        kwargs["target"] = self._create_target(kwargs["target"])

        # TODO: Check target and task
        # get subsample if it needed
        subsample = train_data
        if self.samples is not None and self.samples < subsample.shape[0]:
            subsample = subsample.sample(self.samples, axis=0, random_state=42)

        # infer roles
        for feat in subsample.columns:
            assert isinstance(
                feat, str
            ), "Feature names must be string," " find feature name: {}, with type: {}".format(feat, type(feat))
            if feat in parsed_roles:
                r = parsed_roles[feat]
                # handle datetimes
                if r.name == "Datetime":
                    # try if it's ok to infer date with given params
                    try:
                        _ = pd.to_datetime(
                            subsample[feat],
                            format=r.format,
                            origin=r.origin,
                            unit=r.unit,
                        )
                    except ValueError:
                        raise ValueError("Looks like given datetime parsing params are not correctly defined")

                # replace default category dtype for numeric roles dtype if cat col dtype is numeric
                if r.name == "Category":
                    # default category role
                    cat_role = self._get_default_role_from_str("category")
                    # check if role with dtypes was exactly defined
                    try:
                        flg_default_params = feat in roles["category"]
                    except KeyError:
                        flg_default_params = False

                    if (
                        flg_default_params
                        and not np.issubdtype(cat_role.dtype, np.number)
                        and np.issubdtype(subsample.dtypes[feat], np.number)
                    ):
                        r.dtype = self._get_default_role_from_str("numeric").dtype

            else:
                # if no - infer
                if self._is_ok_feature(subsample[feat]):
                    r = self._guess_role(subsample[feat])

                else:
                    r = DropRole()

            # set back
            if r.name != "Drop":
                self._roles[feat] = r
                self._used_features.append(feat)
            else:
                self._dropped_features.append(feat)

        assert len(self.used_features) > 0, "All features are excluded for some reasons"
        # assert len(self.used_array_attrs) > 0, 'At least target should be defined in train dataset'
        # create folds

        folds = set_sklearn_folds(
            self.task,
            kwargs["target"].values,
            cv=self.cv,
            random_state=self.random_state,
            group=None if "group" not in kwargs else kwargs["group"],
        )
        if folds is not None:
            kwargs["folds"] = Series(folds, index=train_data.index)

        # get dataset
        dataset = PandasDataset(train_data[self.used_features], self.roles, task=self.task, **kwargs)
        if self.advanced_roles:
            new_roles = self.advanced_roles_guess(dataset, manual_roles=parsed_roles)

            droplist = [x for x in new_roles if new_roles[x].name == "Drop" and not self._roles[x].force_input]
            self.upd_used_features(remove=droplist)
            self._roles = {x: new_roles[x] for x in new_roles if x not in droplist}
            dataset = PandasDataset(train_data[self.used_features], self.roles, task=self.task, **kwargs)

        return dataset

    def _create_target(self, target: Union[Series, DataFrame]):
        """Validate target column and create class mapping is needed.

        Args:
            target: Column with target values.

        Returns:
            Transformed target.

        """
        self.class_mapping = None

        if (self.task.name == "binary") or (self.task.name == "multiclass"):
            target, self.class_mapping = self.check_class_target(target)
        elif self.task.name == "multilabel":
            self.class_mapping = {}

            for col in target.columns:
                target_col, class_mapping = self.check_class_target(target[col])
                self.class_mapping[col] = class_mapping
                target.loc[:, col] = target_col.values

            self._n_classes = len(target.columns) * 2
            # TODO: interpretation

        else:
            assert not np.isnan(target.values).any(), "Nan in target detected"
        return target

    def check_class_target(self, target) -> Tuple[pd.Series, Optional[Union[Mapping, Dict[str, Mapping]]]]:
        """Validate target values."""
        target = pd.Series(target)
        cnts = target.value_counts(dropna=False)
        assert np.nan not in cnts.index, "Nan in target detected"
        unqiues = cnts.index.values
        srtd = np.sort(unqiues)
        # WARNING: Multilabel task
        self._n_classes = len(unqiues)

        # if we have multiclass, cv parameter should be less or equal than quantity of samples in the smallest class
        if self.task.name == "multiclass":
            # smallest - quantity of samples in the smallest class
            target_val_counts = target.value_counts(dropna=False)
            smallest = int(target_val_counts.values[-1])
            smallest_class = target_val_counts.index[-1]
            if smallest == 1:
                # some class is represented by 1 sample, so we cannot split dataset correctly
                raise ValueError(
                    f"Class {smallest_class} of target has only 1 sample, dataset cannot be splitted stratified"
                )
            elif smallest < self.cv:
                # we can split dataset correctly, but to do this we change cv parameter for this and warn user about it
                warnings.warn(
                    f"Target class {smallest_class} has only {smallest} samples in the dataset, which is less than {self.cv} (cross-validation parameter), Last one will be changed to {smallest} (not recommended)"
                )
                self.cv = smallest

        # case - target correctly defined and no mapping
        if (np.arange(srtd.shape[0]) == srtd).all():
            assert srtd.shape[0] > 1, "Less than 2 unique values in target"
            if (self.task.name == "binary") or (self.task.name == "multilabel"):
                assert srtd.shape[0] == 2, "Binary task and more than 2 values in target"
            return target, None

        # case - create mapping
        class_mapping = {n: x for (x, n) in enumerate(unqiues)}
        return target.map(class_mapping).astype(np.int32), class_mapping

    def _get_default_role_from_str(self, name) -> RoleType:
        """Get default role for string name according to automl's defaults and user settings.

        Args:
            name: name of role to get.

        Returns:
            role object.

        """
        name = name.lower()
        try:
            role_params = self.roles_params[name]
        except KeyError:
            role_params = {}

        return ColumnRole.from_string(name, **role_params)

    def _guess_role(self, feature: Series) -> RoleType:
        """Try to infer role, simple way.

        If convertible to float -> number.
        Else if convertible to datetime -> datetime.
        Else category.

        Args:
            feature: Column from dataset.

        Returns:
            Feature role.

        """
        # TODO: Plans for advanced roles guessing
        # check if default numeric dtype defined
        num_dtype = self._get_default_role_from_str("numeric").dtype
        # check if feature is number
        try:
            _ = feature.astype(num_dtype)
            return NumericRole(num_dtype)
        except ValueError:
            pass
        except TypeError:
            pass

        # check if default format is defined
        date_format = self._get_default_role_from_str("datetime").format
        # check if it's datetime
        dt_role = DatetimeRole(np.datetime64, date_format=date_format)
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                t = cast(pd.Series, pd.to_datetime(feature, format=date_format))
        except (ValueError, AttributeError, TypeError):
            # else category
            return CategoryRole(object)

        try:
            _ = t.dt.tz_localize("UTC")
            return dt_role
        except TypeError:
            # The exception is raised when TimeSeries is tz-aware and tz is not None
            return dt_role
        except:
            return CategoryRole(object)

    def _is_ok_feature(self, feature) -> bool:
        """Check if column is filled well to be a feature.

        Args:
            feature: Column from dataset.

        Returns:
            ``True`` if nan ratio and frequency are not high.

        """
        if feature.isnull().mean() >= self.max_nan_rate:
            return False
        if (feature.value_counts().values[0] / feature.shape[0]) >= self.max_constant_rate:
            return False
        return True

    def read(self, data: DataFrame, features_names: Any = None, add_array_attrs: bool = False) -> PandasDataset:
        """Read dataset with fitted metadata.

        Args:
            data: Data.
            features_names: Not used.
            add_array_attrs: Additional attributes, like
                target/group/weights/folds.

        Returns:
            Dataset with new columns.

        """
        kwargs = {}
        if add_array_attrs:
            for array_attr in self.used_array_attrs:
                col_name = self.used_array_attrs[array_attr]
                try:
                    val = data.loc[:, col_name]
                except KeyError:
                    continue

                if array_attr == "target" and self.class_mapping is not None:
                    if len(val.shape) == 1:
                        val = Series(
                            val.map(self.class_mapping).values,
                            index=data.index,
                            name=col_name,
                        )
                    else:
                        for col in val.columns:
                            if self.class_mapping[col] is not None:
                                val.loc[:, col] = val.loc[:, col].map(self.class_mapping[col])

                kwargs[array_attr] = val

        dataset = PandasDataset(data[self.used_features], roles=self.roles, task=self.task, **kwargs)

        return dataset

    def advanced_roles_guess(self, dataset: PandasDataset, manual_roles: Optional[RolesDict] = None) -> RolesDict:
        """Advanced roles guess over user's definition and reader's simple guessing.

        Strategy - compute feature's NormalizedGini
        for different encoding ways and calc stats over results.
        Role is inferred by comparing performance stats with manual rules.
        Rule params are params of roles guess in init.
        Defaults are ok in general case.

        Args:
            dataset: Input PandasDataset.
            manual_roles: Dict of user defined roles.

        Returns:
            Dict.

        """
        if manual_roles is None:
            manual_roles = {}
        top_scores = []
        new_roles_dict = dataset.roles
        advanced_roles_params = deepcopy(self.advanced_roles_params)
        drop_co = advanced_roles_params.pop("drop_score_co")

        # guess roles nor numerics
        stat = get_numeric_roles_stat(
            dataset,
            manual_roles=manual_roles,
            random_state=self.random_state,
            subsample=self.samples,
            n_jobs=self.n_jobs,
        )

        if len(stat) > 0:
            # upd stat with rules
            stat = calc_encoding_rules(stat, **advanced_roles_params)
            new_roles_dict = {**new_roles_dict, **rule_based_roles_guess(stat)}
            top_scores.append(stat["max_score"])

        # guess categories handling type
        stat = get_category_roles_stat(
            dataset,
            random_state=self.random_state,
            subsample=self.samples,
            n_jobs=self.n_jobs,
        )
        if len(stat) > 0:
            # upd stat with rules
            # TODO: add sample params
            stat = calc_category_rules(stat)
            new_roles_dict = {**new_roles_dict, **rule_based_cat_handler_guess(stat)}
            top_scores.append(stat["max_score"])

        # get top scores of feature
        if len(top_scores) > 0:
            top_scores = pd.concat(top_scores, axis=0)
            # TODO: Add sample params

            null_scores = get_null_scores(
                dataset,
                list(top_scores.index),
                random_state=self.random_state,
                subsample=self.samples,
            )
            top_scores = pd.concat([null_scores, top_scores], axis=1).max(axis=1)
            rejected = list(top_scores[top_scores < drop_co].index)
            logger.info3("Feats was rejected during automatic roles guess: {0}".format(rejected))
            new_roles_dict = {**new_roles_dict, **{x: DropRole() for x in rejected}}

        return new_roles_dict


class DictToPandasSeqReader(PandasToPandasReader):
    """Reader with sequential support to convert :class:`~pandas.DataFrame` to AutoML's :class:`~lightautoml.dataset.np_pd_dataset.PandasDataset`.

    Stages:

        - Same function for plain dataset as PandasToPandasReader.
        - Parse sequential data (simple auto-typing).

    Args:
        task: Task object.
        samples: Number of elements used when checking role type.
        max_nan_rate: Maximum nan-rate.
        max_constant_rate: Maximum constant rate.
        cv: CV Folds.
        random_state: Random seed.
        roles_params: dict of params of features roles. \
            Ex. {'numeric': {'dtype': np.float32}, 'datetime': {'date_format': '%Y-%m-%d'}}
            It's optional and commonly comes from config
        n_jobs: Int number of processes.
        advanced_roles: Param of roles guess (experimental, do not change).
        numeric_unqiue_rate: Param of roles guess (experimental, do not change).
        max_to_3rd_rate: Param of roles guess (experimental, do not change).
        binning_enc_rate: Param of roles guess (experimental, do not change).
        raw_decr_rate: Param of roles guess (experimental, do not change).
        max_score_rate: Param of roles guess (experimental, do not change).
        abs_score_val: Param of roles guess (experimental, do not change).
        drop_score_co: Param of roles guess (experimental, do not change).
        seq_params: sequence-related params.
        **kwargs: For now not used.

    """

    def __init__(
        self,
        task: Task,
        samples: Optional[int] = 100000,
        max_nan_rate: float = 0.999,
        max_constant_rate: float = 0.999,
        cv: int = 5,
        random_state: int = 42,
        roles_params: Optional[dict] = None,
        n_jobs: int = 4,
        # params for advanced roles guess
        advanced_roles: bool = True,
        numeric_unique_rate: float = 0.999,
        max_to_3rd_rate: float = 1.1,
        binning_enc_rate: float = 2,
        raw_decr_rate: float = 1.1,
        max_score_rate: float = 0.2,
        abs_score_val: float = 0.04,
        drop_score_co: float = 0.01,
        seq_params=None,
        **kwargs: Any,
    ):

        super().__init__(task)
        self.samples = samples
        self.max_nan_rate = max_nan_rate
        self.max_constant_rate = max_constant_rate
        self.cv = cv
        self.random_state = random_state
        self.n_jobs = n_jobs

        self.roles_params = roles_params
        if roles_params is None:
            self.roles_params = {}
        self.target = None
        self.advanced_roles = advanced_roles
        self.advanced_roles_params = {
            "numeric_unique_rate": numeric_unique_rate,
            "max_to_3rd_rate": max_to_3rd_rate,
            "binning_enc_rate": binning_enc_rate,
            "raw_decr_rate": raw_decr_rate,
            "max_score_rate": max_score_rate,
            "abs_score_val": abs_score_val,
            "drop_score_co": drop_score_co,
        }
        if seq_params is not None:
            self.seq_params = seq_params
        else:
            self.seq_params = {
                "seq": {"case": "next_values", "params": {"n_target": 7, "history": 7, "step": 1, "from_last": True}}
            }
        self.params = kwargs

        self.ti = {}
        self.meta = {}

    def create_ids(self, seq_data, plain_data, dataset_name):
        """Calculate ids for different seq tasks."""
        if self.seq_params[dataset_name]["case"] == "next_values":
            self.ti[dataset_name] = TopInd(
                scheme=self.seq_params[dataset_name].get("scheme", None),
                roles=self.meta[dataset_name]["roles"],
                **self.seq_params[dataset_name]["params"],
            )
            self.ti[dataset_name].read(seq_data, plain_data)

        elif self.seq_params[dataset_name]["case"] == "ids":
            self.ti[dataset_name] = IDSInd(
                scheme=self.seq_params[dataset_name].get("scheme", None), **self.seq_params[dataset_name]["params"]
            )
            self.ti[dataset_name].read(seq_data, plain_data)

    def parse_seq(self, seq_dataset, plain_data, dataset_name, parsed_roles, roles):
        """Method to read sequential data."""
        subsample = seq_dataset
        if self.samples is not None and self.samples < subsample.shape[0]:
            subsample = subsample.sample(self.samples, axis=0, random_state=42)

        seq_roles = {}
        seq_features = []
        kwargs = {}
        used_array_attrs = {}
        for feat in seq_dataset.columns:
            assert isinstance(
                feat, str
            ), "Feature names must be string," " find feature name: {}, with type: {}".format(feat, type(feat))

            if feat in parsed_roles:
                r = parsed_roles[feat]
                if isinstance(r, str):
                    # get default role params if defined
                    r = self._get_default_role_from_str(r)
                # handle datetimes

                if r.name == "Datetime" or r.name == "Date":
                    # try if it's ok to infer date with given params
                    try:
                        _ = pd.to_datetime(subsample[feat], format=r.format, origin=r.origin, unit=r.unit)
                    except ValueError:
                        raise ValueError("Looks like given datetime parsing params are not correctly defined")

                # replace default category dtype for numeric roles dtype if cat col dtype is numeric
                if r.name == "Category":
                    # default category role
                    cat_role = self._get_default_role_from_str("category")
                    # check if role with dtypes was exactly defined
                    try:
                        flg_default_params = feat in roles["category"]
                    except KeyError:
                        flg_default_params = False

                    if (
                        flg_default_params
                        and not np.issubdtype(cat_role.dtype, np.number)
                        and np.issubdtype(subsample.dtypes[feat], np.number)
                    ):
                        r.dtype = self._get_default_role_from_str("numeric").dtype

                if r.name == "Target":
                    r = self._get_default_role_from_str("numeric")

            else:
                # if no - infer
                if self._is_ok_feature(subsample[feat]):
                    r = self._guess_role(subsample[feat])

                else:
                    r = DropRole()

            parsed_roles[feat] = r

            if r.name in attrs_dict:
                if attrs_dict[r.name] in ["target"]:
                    pass
                else:
                    kwargs[attrs_dict[r.name]] = seq_dataset[feat]
                    used_array_attrs[attrs_dict[r.name]] = feat
                    r = DropRole()

            # set back
            if r.name != "Drop":
                self._roles[feat] = r
                self._used_features.append(feat)
                seq_roles[feat] = r
                seq_features.append(feat)
            else:
                self._dropped_features.append(feat)

        assert len(seq_features) > 0, "All features are excluded for some reasons"
        self.meta[dataset_name] = {}
        self.meta[dataset_name]["roles"] = seq_roles
        self.meta[dataset_name]["features"] = seq_features
        self.meta[dataset_name]["attributes"] = used_array_attrs

        self.create_ids(seq_dataset, plain_data, dataset_name)
        self.meta[dataset_name].update(
            **{
                "seq_idx_data": self.ti[dataset_name].create_data(seq_dataset, plain_data=plain_data),
                "seq_idx_target": self.ti[dataset_name].create_target(seq_dataset, plain_data=plain_data),
            }
        )

        if self.meta[dataset_name]["seq_idx_target"] is not None:
            assert len(self.meta[dataset_name]["seq_idx_data"]) == len(
                self.meta[dataset_name]["seq_idx_target"]
            ), "Time series ids don`t match"

        seq_dataset = SeqNumpyPandasDataset(
            data=seq_dataset[self.meta[dataset_name]["features"]],
            features=self.meta[dataset_name]["features"],
            roles=self.meta[dataset_name]["roles"],
            idx=self.meta[dataset_name]["seq_idx_target"]
            if self.meta[dataset_name]["seq_idx_target"] is not None
            else self.meta[dataset_name]["seq_idx_data"],
            name=dataset_name,
            scheme=self.seq_params[dataset_name].get("scheme", None),
            **kwargs,
        )
        return seq_dataset, parsed_roles

    def fit_read(
        self, train_data: Dict, features_names: Any = None, roles: UserDefinedRolesDict = None, **kwargs: Any
    ) -> PandasDataset:
        """Get dataset with initial feature selection.

        Args:
            train_data: Input data in dict format.
            features_names: Ignored. Just to keep signature.
            roles: Dict of features roles in format ``{RoleX: ['feat0', 'feat1', ...], RoleY: 'TARGET', ....}``.
            **kwargs: Can be used for target/group/weights.

        Returns:
            Dataset with selected features.

        """
        # logger.info('Train data shape: {}'.format(train_data.shape))

        if roles is None:
            roles = {}
        # transform roles from user format {RoleX: ['feat0', 'feat1', ...], RoleY: 'TARGET', ....}
        # to automl format {'feat0': RoleX, 'feat1': RoleX, 'TARGET': RoleY, ...}

        plain_data, seq_data = train_data.get("plain", None), train_data.get("seq", None)
        plain_features = set(plain_data.columns) if plain_data is not None else {}
        parsed_roles = roles_parser(roles)
        # transform str role definition to automl ColumnRole

        seq_datasets = {}
        if seq_data is not None:
            for dataset_name, dataset in seq_data.items():
                seq_dataset, parsed_roles = self.parse_seq(dataset, plain_data, dataset_name, parsed_roles, roles)
                seq_datasets[dataset_name] = seq_dataset
        else:
            seq_datasets = None

        for feat in parsed_roles:
            r = parsed_roles[feat]
            if isinstance(r, str):
                # get default role params if defined
                r = self._get_default_role_from_str(r)

            # check if column is defined like target/group/weight etc ...
            if feat in plain_features:
                if r.name in attrs_dict:
                    # defined in kwargs is rewritten.. TODO: Maybe raise warning if rewritten?

                    if ((self.task.name == "multi:reg") or (self.task.name == "multilabel")) and (
                        attrs_dict[r.name] == "target"
                    ):
                        if attrs_dict[r.name] in kwargs:
                            kwargs[attrs_dict[r.name]].append(feat)
                            self._used_array_attrs[attrs_dict[r.name]].append(feat)
                        else:
                            kwargs[attrs_dict[r.name]] = [feat]
                            self._used_array_attrs[attrs_dict[r.name]] = [feat]
                    else:
                        self._used_array_attrs[attrs_dict[r.name]] = feat
                        kwargs[attrs_dict[r.name]] = plain_data.loc[:, feat]
                    r = DropRole()

                # add new role
                parsed_roles[feat] = r
        # add target from seq dataset to plain
        for seq_name, values in self.meta.items():
            if values["seq_idx_target"] is not None:
                kwargs["target"] = pd.DataFrame(
                    seq_datasets[seq_name].to_sequence((slice(None), roles["target"])).data[:, :, 0].astype(float)
                )
                break

        assert "target" in kwargs, "Target should be defined"
        if isinstance(kwargs["target"], list):
            kwargs["target"] = plain_data.loc[:, kwargs["target"]]

        self.target = kwargs["target"].name if isinstance(kwargs["target"], pd.Series) else kwargs["target"].columns
        kwargs["target"] = self._create_target(kwargs["target"])

        # TODO: Check target and task
        # get subsample if it needed
        if plain_data is not None:

            subsample = plain_data
            if self.samples is not None and self.samples < subsample.shape[0]:
                subsample = subsample.sample(self.samples, axis=0, random_state=42)

            # infer roles
            for feat in subsample.columns:
                assert isinstance(
                    feat, str
                ), "Feature names must be string," " find feature name: {}, with type: {}".format(feat, type(feat))
                if feat in parsed_roles:
                    r = parsed_roles[feat]
                    # handle datetimes
                    if r.name == "Datetime":
                        # try if it's ok to infer date with given params
                        try:
                            _ = pd.to_datetime(subsample[feat], format=r.format, origin=r.origin, unit=r.unit)
                        except ValueError:
                            raise ValueError("Looks like given datetime parsing params are not correctly defined")

                    # replace default category dtype for numeric roles dtype if cat col dtype is numeric
                    if r.name == "Category":
                        # default category role
                        cat_role = self._get_default_role_from_str("category")
                        # check if role with dtypes was exactly defined
                        try:
                            flg_default_params = feat in roles["category"]
                        except KeyError:
                            flg_default_params = False

                        if (
                            flg_default_params
                            and not np.issubdtype(cat_role.dtype, np.number)
                            and np.issubdtype(subsample.dtypes[feat], np.number)
                        ):
                            r.dtype = self._get_default_role_from_str("numeric").dtype

                else:
                    # if no - infer
                    if self._is_ok_feature(subsample[feat]):
                        r = self._guess_role(subsample[feat])

                    else:
                        r = DropRole()

                # set back
                if r.name != "Drop":
                    self._roles[feat] = r
                    self._used_features.append(feat)
                else:
                    self._dropped_features.append(feat)

            assert (
                len(set(self.used_features) & set(subsample.columns)) > 0
            ), "All features are excluded for some reasons"

        if self.cv is not None:
            folds = set_sklearn_folds(
                self.task,
                kwargs["target"],
                cv=self.cv,
                random_state=self.random_state,
                group=None if "group" not in kwargs else kwargs["group"],
            )
            kwargs["folds"] = Series(
                folds, index=plain_data.index if plain_data is not None else np.arange(len(kwargs["target"]))
            )

        # get dataset
        self.plain_used_features = sorted(list(set(self.used_features) & set(plain_features)))
        self.plain_roles = {key: value for key, value in self._roles.items() if key in set(self.plain_used_features)}

        dataset = PandasDataset(
            plain_data[self.plain_used_features] if plain_data is not None else pd.DataFrame(),
            self.plain_roles,
            task=self.task,
            **kwargs,
        )
        if self.advanced_roles:
            new_roles = self.advanced_roles_guess(dataset, manual_roles=parsed_roles)
            droplist = [x for x in new_roles if new_roles[x].name == "Drop" and not self._roles[x].force_input]
            self.upd_used_features(remove=droplist)
            self._roles = {x: new_roles[x] for x in new_roles if x not in droplist}
            self.plain_used_features = sorted(list(set(self.used_features) & set(plain_features)))
            self.plain_roles = {
                key: value for key, value in self._roles.items() if key in set(self.plain_used_features)
            }
            dataset = PandasDataset(
                plain_data[self.plain_used_features] if plain_data is not None else pd.DataFrame(),
                self.plain_roles,
                task=self.task,
                **kwargs,
            )

        for seq_name, values in self.meta.items():
            seq_datasets[seq_name].idx = values["seq_idx_data"]

        dataset.seq_data = seq_datasets

        return dataset

    def read(self, data, features_names: Any = None, add_array_attrs: bool = False) -> PandasDataset:
        """Read dataset with fitted metadata.

        Args:
            data: Data.
            features_names: Not used.
            add_array_attrs: Additional attributes, like target/group/weights/folds.

        Returns:
            Dataset with new columns.

        """
        plain_data, seq_data = data.get("plain", None), data.get("seq", None)
        seq_datasets = {}
        if seq_data is not None:
            for dataset_name, dataset in seq_data.items():
                test_idx = self.ti[dataset_name].create_test(dataset, plain_data=plain_data)
                kwargs = {}
                columns = set(dataset.columns)
                for role, col in self.meta[dataset_name]["attributes"].items():
                    if col in columns:
                        kwargs[role] = dataset[col]

                seq_dataset = SeqNumpyPandasDataset(
                    data=dataset[self.meta[dataset_name]["features"]],
                    features=self.meta[dataset_name]["features"],
                    roles=self.meta[dataset_name]["roles"],
                    idx=test_idx,
                    name=dataset_name,
                    scheme=self.seq_params[dataset_name].get("scheme", None),
                    **kwargs,
                )

                seq_datasets[dataset_name] = seq_dataset
        else:
            seq_datasets = None

        kwargs = {}
        if add_array_attrs:
            for array_attr in self.used_array_attrs:
                col_name = self.used_array_attrs[array_attr]
                try:
                    val = plain_data[col_name]
                except KeyError:
                    continue

                if array_attr == "target" and self.class_mapping is not None:
                    if len(val.shape) == 1:
                        val = Series(val.map(self.class_mapping).values, index=plain_data.index, name=col_name)
                    else:
                        for col in val.columns:
                            if self.class_mapping[col] is not None:
                                val.loc[:, col] = val.loc[:, col].map(self.class_mapping[col])
                kwargs[array_attr] = val

        dataset = PandasDataset(
            plain_data[self.plain_used_features] if plain_data is not None else pd.DataFrame(),
            self.plain_roles,
            task=self.task,
            **kwargs,
        )

        dataset.seq_data = seq_datasets
        return dataset



"""Generate sequential features."""

from typing import List
from typing import Union

import numpy as np

from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import NumericRole
from ..dataset.seq_np_pd_dataset import SeqNumpyPandasDataset
from ..transformers.base import LAMLTransformer


# type - something that can be converted to pandas dataset
NumpyTransformable = Union[NumpyDataset, PandasDataset]
NumpyCSR = Union[NumpyDataset, CSRSparseDataset]


class SeqLagTransformer(LAMLTransformer):
    """LAG.

    Flattens different features

    Args:
        lags: int (number of lags to compute), list / np.ndarray (numbers of certain lags to compute).

    """

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "lag"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    @staticmethod
    def _get_attributes(dataset):
        params = {}
        for attribute in dataset._array_like_attrs:
            _data = []
            _d = getattr(dataset, attribute).values
            for row in np.arange(len(dataset)):
                _data.append(_d[dataset.idx[row]][-1])
            _data = np.array(_data)
            params[attribute] = _data
        return params

    def __init__(self, lags: Union[int, List[int], np.ndarray] = 10):
        if isinstance(lags, list):
            self.lags = np.array(lags)
        if isinstance(lags, int):
            self.lags = np.arange(lags)

    def fit(self, dataset):
        """Fit algorithm on seq dataset.

        Args:
            dataset: NumpyDataset.

        Returns:
            Fitted transformer.

        """
        sample_data = dataset.to_sequence([0]).data  # (1 observation, history, features)

        # convert to accepted dtype and get attributes
        # leave only correct lags (less than number of observations in sample_data)
        self.current_correct_lags = self.lags.copy()[self.lags < sample_data.data.shape[1]]

        feats = []
        for feat in dataset.features:
            feats.extend([self._fname_prefix + f"_{i}" + "__" + feat for i in reversed(self.current_correct_lags)])
        self._features = list(feats)
        return self

    def transform(self, dataset) -> NumpyDataset:
        """Transform input seq dataset to normal numpy representation.

        Args:
            dataset: seq.

        Returns:
            Numpy dataset with lag features.

        """
        # checks here
        super().transform(dataset)

        # convert to accepted dtype and get attributes
        data_seq = dataset.to_sequence().data
        data = data_seq[:, (data_seq.shape[1] - 1) - self.current_correct_lags[::-1], :]

        params = self._get_attributes(dataset)

        # transform
        data = np.moveaxis(data, 1, 2).reshape(len(data), -1)

        # create resulted
        return NumpyDataset(data, self.features, NumericRole(np.float32), **params)


class SeqDiffTransformer(SeqLagTransformer, LAMLTransformer):
    """Diff.

    Args:
        diffs: int (number of diffs to compute), list / np.ndarray (numbers of certain diffs to compute).
        diffs = 0 means no diff (y_t), 1 means diff between last and previous (y_t-y_{t-1}), 2 means diff between last and previous to previous (y_t-t_{t-2}), etc.

    """

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "diff"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(self, diffs: Union[int, List[int], np.ndarray] = 10, flag_del_0_diff=False):
        SeqLagTransformer.__init__(self, lags=diffs)
        self.flag_del_0_diff = (
            flag_del_0_diff  # if True, we need to drop diff with number 0 to avoid column duplication
        )

    def fit(self, dataset):
        """Fit algorithm on seq dataset.

        Args:
            dataset: NumpyDataset.

        Returns:
            Fitted transformer.

        """
        if self.flag_del_0_diff:
            self.lags = self.lags[self.lags > 0]  # drop diff=0
        return SeqLagTransformer.fit(self, dataset)

    def transform(self, dataset) -> NumpyDataset:
        """Transform input seq dataset to normal numpy representation.

        Args:
            dataset: seq.

        Returns:
            Numpy dataset with lag features.

        """
        # checks here
        LAMLTransformer.transform(self, dataset)

        # convert to accepted dtype and get attributes
        data_seq = dataset.to_sequence().data
        data_seq_t = data_seq[:, (data_seq.shape[1] - 1) - np.array([0]), :]

        params = self._get_attributes(dataset)

        # transform
        if 0 in self.current_correct_lags:
            data_seq_diffs = data_seq[:, (data_seq.shape[1] - 1) - self.current_correct_lags[::-1][:-1], :]
            data_seq_diffs = np.concatenate(
                (data_seq_diffs, np.zeros((data_seq_diffs.shape[0], 1, data_seq_diffs.shape[2]))), axis=1
            )
        else:
            data_seq_diffs = data_seq[:, (data_seq.shape[1] - 1) - self.current_correct_lags[::-1], :]

        data_t = np.moveaxis(data_seq_t, 1, 2)[:, :, ::-1].reshape(len(data_seq_t), -1)
        data_t_diffs = np.moveaxis(data_seq_diffs, 1, 2).reshape(len(data_seq_diffs), -1)
        final_t = np.repeat(data_t, len(self.current_correct_lags), axis=1) - data_t_diffs

        # create resulted
        return NumpyDataset(final_t, self.features, NumericRole(np.float32), **params)


class SeqNumCountsTransformer(LAMLTransformer):
    """NC."""

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "numcount"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(self):
        pass

    def fit(self, dataset):
        """Fit algorithm on seq dataset.

        Args:
            dataset: NumpyDataset.

        Returns:
            Self.

        """
        # set transformer names and add checks
        # for check_func in self._fit_checks:
        #    check_func(dataset)
        # set transformer features

        feats = [self._fname_prefix + "__" + dataset.name]

        self._features = list(feats)
        return self

    def transform(self, dataset) -> NumpyDataset:
        """Transform input seq dataset to normal numpy representation.

        Args:
            dataset: seq.

        Returns:
            Numpy dataset with len of the sequence.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        data = dataset.apply_func((slice(None)), len).reshape(-1, 1)
        # transform

        # print('name', dataset.name)
        # print('scheme', dataset.scheme)
        # create resulted
        return NumpyDataset(data, self.features, NumericRole(np.float32))


class SeqStatisticsTransformer(LAMLTransformer):
    """SSF."""

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "stat"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(self):
        pass

    def fit(self, dataset):
        """Fit algorithm on seq dataset.

        Args:
            dataset: NumpyDataset.

        Returns:
            Self.

        """
        feats = []
        for feat in dataset.features:
            feats.extend([self._fname_prefix + "__" + dataset.name + "__" + feat])
        self._features = list(feats)
        return self

    def transform(self, dataset) -> NumpyDataset:
        """Transform input seq dataset to normal numpy representation.

        Args:
            dataset: seq.

        Returns:
            Numpy dataset with last known feature in the sequence.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        data = dataset.apply_func((slice(None)), self._get_last)

        return NumpyDataset(data, self.features, NumericRole(np.float32))

    @staticmethod
    def _std(x):
        return np.std(x, axis=0)

    @staticmethod
    def _get_last(x):
        if len(x) > 0:
            return x[-1, :]
        else:
            return np.array([np.NaN] * x.shape[-1])


class GetSeqTransformer(LAMLTransformer):
    """LAG."""

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "seq"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(self, name="seq"):

        self.name = name

    def fit(self, dataset):
        """Fit algorithm on seq dataset.

        Args:
            dataset: NumpyDataset.

        Returns:
            Self.

        """
        # set transformer names and add checks
        # for check_func in self._fit_checks:
        #    check_func(dataset)
        # set transformer features
        sample_data = dataset.seq_data.get(self.name)
        # convert to accepted dtype and get attributes

        feats = []
        self.roles = {}
        for feat, role in sample_data.roles.items():
            # feats.extend([self._fname_prefix + '__' + feat])
            feats.extend([feat])
            self.roles[feats[-1]] = role
        self._features = list(feats)
        return self

    def transform(self, dataset) -> NumpyDataset:
        """Transform input seq dataset to normal numpy representation.

        Args:
            dataset: seq.

        Returns:
            Numpy dataset with lag features.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes

        dataset = dataset.seq_data.get(self.name)
        data = dataset.data
        data.columns = self._features
        kwargs = {}
        for attr in dataset._array_like_attrs:
            kwargs[attr] = dataset.__dict__[attr]

        # create resulted
        result = SeqNumpyPandasDataset(
            data=data,
            features=self.features,
            roles=self.roles,
            idx=dataset.idx,
            name=dataset.name,
            scheme=dataset.scheme,
            **kwargs,
        )

        return result
